from logging import getLogger

from ..worker import run
from ...commands.args import WEB, UPLOAD
from ...commands.upload import STREAM, NAME, upload_files
from ...lib.log import Record

log = getLogger(__name__)


class Upload:

    def __init__(self, config):
        self.__config = config

    def __call__(self, request, s):
        files = [{NAME: file.filename, STREAM: file.stream} for file in request.files.getlist('files')]
        items = request.form.getlist('kit')
        # we do this in two stages
        # first, immediate saving of files while web browser waiting for response
        upload_files(Record(log), self.__config, files=files, items=items)
        # second, start rest of ingest process in background
        run(self.__config, UPLOAD, Upload, f'{WEB}-{UPLOAD}.log')



from logging import getLogger

from scipy import optimize

log = getLogger(__name__)


def fit(params, data, model, cost, vary=None, strategy=None, bounds=None, **kargs):

    if vary is None: vary = list(params.keys())
    if strategy is None: strategy = immediate

    for to_vary in strategy(vary):
        log.info(f'Varying {",".join(to_vary)}')

        def func(x):
            for (name, value) in zip(to_vary, x):
                params[name] = value
            return cost(model(params, data))

        x0 = [params[name] for name in to_vary]
        bds = [bounds.get(name, (None, None)) for name in to_vary] if bounds else None
        result = optimize.minimize(func, x0, bounds=bds, method='Powell', **kargs)

        if not result.success:
            log.warning(f'Optimize failed: {result.message}')

        for (name, value) in zip(to_vary, result.x):
            log.info(f'{name} = {value}')
            params[name] = value

    return params


def immediate(vary):
    yield vary


def progressive(vary):
    to_vary = []
    for name in vary:
        to_vary.append(name)
        yield to_vary



from .utils import ProcessCalculator, RerunWhenNewActivitiesMixin
from ...data.cluster import hulls_from_last_activity, sectors_from_hulls
from ...sql import Timestamp
from ...sql.tables.sector import DEFAULT_GROUP_RADIUS_KM


class ClusterCalculator(RerunWhenNewActivitiesMixin, ProcessCalculator):

    def __init__(self, *args, excess=0.1, radius_km=DEFAULT_GROUP_RADIUS_KM, **kargs):
        super().__init__(*args, excess=excess, **kargs)
        self.__radius_km = radius_km

    def _run_one(self, missed):
        with self._config.db.session_context() as s:
            with Timestamp(owner=self.owner_out).on_success(s):
                sector_group = hulls_from_last_activity(s, self.__radius_km)
                if sector_group:
                    sectors_from_hulls(s, sector_group)






from tempfile import TemporaryDirectory

from sqlalchemy.sql.functions import count

from ch2.commands.args import V, DEV, FORCE, bootstrap_db, BASE
from ch2.commands.constants import constants
from ch2.commands.upload import upload
from ch2.common.args import mm, m
from ch2.config.profiles.default import default
from ch2.names import N
from ch2.pipeline.pipeline import run_pipeline
from ch2.sql.tables.activity import ActivityJournal
from ch2.sql.tables.pipeline import PipelineType
from ch2.sql.tables.statistic import StatisticJournal, StatisticJournalFloat, StatisticName
from tests import LogTestCase, random_test_user


class TestActivities(LogTestCase):

    def test_activities(self):

        user = random_test_user()
        bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        config = bootstrap_db(user, m(V), '5', 'constants', 'set', 'SRTM1.dir',
                              '~/.ch2/permanent/srtm1', mm(FORCE))
        constants(config)

        with TemporaryDirectory() as f:
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), 'upload',
                                  'data/test/source/personal/2018-08-27-rec.fit')
            upload(config)

        with config.db.session_context() as s:
            n_raw = s.query(count(StatisticJournalFloat.id)). \
                join(StatisticName). \
                filter(StatisticName.name == N.SRTM1_ELEVATION).scalar()
            self.assertEqual(2099, n_raw)
            n_fix = s.query(count(StatisticJournalFloat.id)). \
                join(StatisticName). \
                filter(StatisticName.name == N.ELEVATION).scalar()
            self.assertEqual(2099, n_fix)
            # WHY does this jump around?
            n = s.query(count(StatisticJournal.id)).scalar()
            # self.assertEqual(62525, n)
            self.assertTrue(n > 30000)
            self.assertTrue(n < 100000)
            journal = s.query(ActivityJournal).one()
            self.assertNotEqual(journal.start, journal.finish)
            print(n)

    def test_segment_bug(self):
        user = random_test_user()
        with TemporaryDirectory() as f:
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
            paths = ['/home/andrew/archive/fit/bike/cotic/2016-07-27-pm-z4.fit']
            run_pipeline(config, PipelineType.PROCESS, paths=paths, force=True)

    def __assert_basic_stats(self, s):
        for name in [N.ACTIVE_DISTANCE, N.ACTIVE_TIME]:
            count = 0
            for stat in s.query(StatisticJournal). \
                    join(StatisticName). \
                    filter(StatisticName.name == name).all():
                count += 1
                print(f'{name} = {stat}')
                self.assertTrue(stat, f'No value for {name}')
            self.assertTrue(count > 0)

    def test_michael(self):
        user = random_test_user()
        with TemporaryDirectory() as f:
            bootstrap_db(user, mm(BASE), f, m(V), '5')
            bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), 'upload',
                                      'data/test/source/other/2019-05-09-051352-Running-iWatchSeries3.fit')
            upload(config)
            with config.db.session_context() as s:
                self.__assert_basic_stats(s)

    def test_heart_alarms(self):
        user = random_test_user()
        with TemporaryDirectory() as f:
            bootstrap_db(user, mm(BASE), f, m(V), '5')
            bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), 'upload',
                                      'data/test/source/personal/2016-07-19-mpu-s-z2.fit')
            upload(config)
            with config.db.session_context() as s:
                for stat in s.query(StatisticJournal). \
                        join(StatisticName). \
                        filter(StatisticName.name == N.ACTIVE_DISTANCE).all():
                    self.assertGreater(stat.value, 30)

    def test_920(self):
        for src in '920xt-2019-05-16_19-42-54.fit', '920xt-2019-05-16_19-42-54.fit':
            user = random_test_user()
            with TemporaryDirectory() as f:
                bootstrap_db(user, mm(BASE), f, m(V), '5')
                bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
                config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), 'upload',
                                               f'data/test/source/other/{src}')
                upload(config)
                with config.db.session_context() as s:
                    self.__assert_basic_stats(s)



from geoalchemy2 import Geometry
from sqlalchemy import Column, Integer, Text, ForeignKey, UniqueConstraint, Index, Float

from ..support import Base


class ClusterInputScratch(Base):

    __tablename__ = 'cluster_input_scratch'

    id = Column(Integer, primary_key=True)
    sector_group_id = Column(Integer, ForeignKey('sector_group.id', ondelete='cascade'), nullable=False)
    geom = Column(Geometry(geometry_type='GeometryM', dimension=3), nullable=False)  # some fragment of a route
    level = Column(Integer)
    group = Column(Integer)
    Index('natural_cluster_input_scratch_ix', 'sector_group_id', 'level', 'group')


class ClusterHull(Base):

    __tablename__ = 'cluster_hull'

    id = Column(Integer, primary_key=True)
    sector_group_id = Column(Integer, ForeignKey('sector_group.id', ondelete='cascade'), nullable=False)
    hull = Column(Geometry(), nullable=False)
    group = Column(Integer)
    level = Column(Integer)
    UniqueConstraint(sector_group_id, level, group)


class ClusterFragmentScratch(Base):

    __tablename__ = 'cluster_fragment_scratch'

    id = Column(Integer, primary_key=True)
    cluster_hull_id = Column(Integer, ForeignKey('cluster_hull.id', ondelete='cascade'), nullable=False, index=True)
    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'), nullable=False, index=True)
    fragment = Column(Geometry('LineStringM'), nullable=False)
    length = Column(Float, nullable=False)
    Index('cluster_fragment_scratch_access_ix', cluster_hull_id, activity_journal_id, length)



from logging import getLogger

import pandas as pd
from geoalchemy2 import WKBElement
from geoalchemy2.shape import to_shape
from sqlalchemy import text, desc

from . import ContentType
from ...common.date import time_to_local_time
from ...data.sector import add_start_finish
from ...diary.model import DB
from ...names import N
from ...pipeline.process import run_pipeline
from ...sql import ActivityJournal, PipelineType, SectorJournal
from ...sql.tables.sector import SectorGroup, DEFAULT_GROUP_RADIUS_KM, SectorType
from ...sql.types import linestringxy
from ...sql.utils import WGS84_SRID, add

log = getLogger(__name__)


class UserDefined:
    '''
    Owner for user defined sectors.
    '''
    pass


class Sector(ContentType):

    def __init__(self, config):
        super().__init__()
        self.__config = config

    def create_sector(self, request, s):
        from ...sql import Sector
        data = request.json
        activity = data['activity']
        activity_journal = s.query(ActivityJournal).filter(ActivityJournal.id == activity).one()
        sector_group = SectorGroup.add(s, to_shape(activity_journal.centre).coords[0],
                                       DEFAULT_GROUP_RADIUS_KM,
                                       f'From activity on {activity_journal.start}')
        full_route = to_shape(self._read_activity_route_wkb(s, activity))
        new_route = full_route.coords[data['start']:data['finish']]
        ls_route = linestringxy(new_route)
        srid_route = f'st_transform(st_setsrid({ls_route}, {WGS84_SRID}), {sector_group.srid})'
        sector = add(s, Sector(type=SectorType.SECTOR, sector_group=sector_group, route=text(srid_route),
                               title=data['name'], owner=UserDefined))
        s.flush()
        add_start_finish(s, sector.id)
        s.commit()
        run_pipeline(self.__config, PipelineType.SECTOR, new_sector_id=sector.id)
        return {'sector': sector.id}

    def _read_activity_route_wkb(self, s, activity_journal_id):
        q = text(f'''
select st_force2d(aj.route_et::geometry)
  from activity_journal as aj
 where aj.id = :activity_journal_id''')
        return WKBElement(s.connection().execute(q, activity_journal_id=activity_journal_id).fetchone()[0])

    def read_sector_journals(self, request, s, sector):
        from ...sql import Sector
        instance = s.query(Sector).filter(Sector.id == sector).one()
        return {'name': instance.title,
                'sector_journals': list(self.__read_sector_journals(s, sector))}

    def __read_sector_journals(self, s, sector):
        '''
        the sqlalchemy below expands into this, which gives a 5s query:

SELECT statistic_journal_text.value AS statistic_journal_text_value
  FROM statistic_journal
  JOIN statistic_journal_text ON statistic_journal.id = statistic_journal_text.id
  JOIN statistic_name ON statistic_name.id = statistic_journal.statistic_name_id
  JOIN activity_topic_field ON activity_topic_field.statistic_name_id = statistic_name.id
  JOIN activity_topic ON activity_topic.id = activity_topic_field.activity_topic_id
  JOIN (source JOIN activity_topic_journal ON source.id = activity_topic_journal.id) ON activity_topic_journal.id = statistic_journal.source_id
  JOIN (source AS source_1 JOIN activity_journal AS activity_journal_1 ON source_1.id = activity_journal_1.id) ON activity_journal_1.file_hash_id = activity_topic_journal.file_hash_id
 WHERE statistic_name.name = 'name'
   AND activity_journal_1.id = 9835
   AND activity_topic.activity_group_id = source_1.activity_group_id;

dropping the reference to source gives this, which is instantaneous.  i do not understand why.

SELECT statistic_journal_text.value AS statistic_journal_text_value
  FROM statistic_journal
  JOIN statistic_journal_text ON statistic_journal.id = statistic_journal_text.id
  JOIN statistic_name ON statistic_name.id = statistic_journal.statistic_name_id
  JOIN activity_topic_field ON activity_topic_field.statistic_name_id = statistic_name.id
  JOIN activity_topic ON activity_topic.id = activity_topic_field.activity_topic_id
  JOIN activity_topic_journal ON activity_topic_journal.id = statistic_journal.source_id
  JOIN (source AS source_1 JOIN activity_journal AS activity_journal_1 ON source_1.id = activity_journal_1.id) ON activity_journal_1.file_hash_id = activity_topic_journal.file_hash_id
 WHERE statistic_name.name = 'name'
   AND activity_journal_1.id = 9835
   AND activity_topic.activity_group_id = source_1.activity_group_id;

        '''
        for sjournal in s.query(SectorJournal). \
                filter(SectorJournal.sector_id == sector). \
                join(ActivityJournal, SectorJournal.activity_journal_id == ActivityJournal.id). \
                order_by(desc(ActivityJournal.start)).all():
            # q = s.query(StatisticJournalText.value). \
            #     join(StatisticName, StatisticName.id == StatisticJournalText.statistic_name_id). \
            #     join(ActivityTopicField, ActivityTopicField.statistic_name_id == StatisticName.id). \
            #     join(ActivityTopic, ActivityTopic.id == ActivityTopicField.activity_topic_id). \
            #     join(ActivityTopicJournal, ActivityTopicJournal.id == StatisticJournal.source_id). \
            #     join(ActivityJournal, ActivityJournal.file_hash_id == ActivityTopicJournal.file_hash_id). \
            #     filter(StatisticName.name == 'name',
            #            ActivityJournal.id == sjournal.activity_journal.id,
            #            ActivityTopic.activity_group_id == ActivityJournal.activity_group_id)
            # activity_name = q.one()[0]
            q = text('''
SELECT statistic_journal_text.value AS statistic_journal_text_value
  FROM statistic_journal
  JOIN statistic_journal_text ON statistic_journal.id = statistic_journal_text.id
  JOIN statistic_name ON statistic_name.id = statistic_journal.statistic_name_id
  JOIN activity_topic_field ON activity_topic_field.statistic_name_id = statistic_name.id
  JOIN activity_topic ON activity_topic.id = activity_topic_field.activity_topic_id
  JOIN activity_topic_journal ON activity_topic_journal.id = statistic_journal.source_id
  JOIN (source AS source_1 JOIN activity_journal AS activity_journal_1 ON source_1.id = activity_journal_1.id) ON activity_journal_1.file_hash_id = activity_topic_journal.file_hash_id
 WHERE statistic_name.name = 'name'
   AND activity_journal_1.id = :activity_journal_id
   AND activity_topic.activity_group_id = source_1.activity_group_id;
''')
            activity_name = s.connection(). \
                execute(q, activity_journal_id=sjournal.activity_journal.id). \
                fetchone()[0]
            yield {DB: sjournal.id,
                   'date': time_to_local_time(sjournal.activity_journal.start),
                   'name': activity_name,
                   'activity_group': sjournal.activity_journal.activity_group.title,
                   'activity_id': sjournal.activity_journal.id,
                   'distance': sjournal.finish_distance - sjournal.start_distance,
                   'time': (sjournal.finish_time - sjournal.start_time).total_seconds(),
                   'elevation': sjournal.finish_elevation - sjournal.start_elevation,
                   'edt': self._read_sector_journal_edt(s, sjournal.id)}

    def _read_sector_journal_edt(self, s, sector_journal_id):
        df = self._read_clipped_d_et(s, sector_journal_id)
        return {
            'elevation': df[N.ELEVATION].tolist(),
            'distance': (df[N.DISTANCE] - df[N.DISTANCE].iloc[0]).tolist(),
            'time': (df[N.ELAPSED_TIME] - df[N.ELAPSED_TIME].iloc[0]).tolist()}

    def _read_clipped_d_et(self, s, sector_journal_id):
        # cannot use route_edt because we need to substring / interpolate
        sql = text(f'''
  with points as (select st_dumppoints(
                            st_linesubstring(
                              aj.route_d::geometry, sj.start_fraction, sj.finish_fraction)) as point
                    from activity_journal as aj,
                         sector_journal as sj
                   where sj.id = :sector_journal_id
                     and aj.id = sj.activity_journal_id)
select st_x((point).geom) as x, st_y((point).geom) as y, st_m((point).geom) as {N.DISTANCE}
  from points;
''')
        log.debug(sql)
        df_d = pd.read_sql(sql, s.connection(),
                           params={'sector_journal_id': sector_journal_id})
        sql = text(f'''
  with points as (select st_dumppoints(
                            st_linesubstring(
                              aj.route_et::geometry, sj.start_fraction, sj.finish_fraction)) as point
                    from activity_journal as aj,
                         sector_journal as sj
                   where sj.id = :sector_journal_id
                     and aj.id = sj.activity_journal_id)
select st_x((point).geom) as x, st_y((point).geom) as y, st_z((point).geom) as {N.ELEVATION},
       st_m((point).geom) as "{N.ELAPSED_TIME}"
  from points;
''')
        log.debug(sql)
        df_et = pd.read_sql(sql, s.connection(),
                            params={'sector_journal_id': sector_journal_id})
        df = pd.merge(df_et, df_d, how='left', left_on=['x', 'y'], right_on=['x', 'y']).dropna()
        return df




from logging import getLogger

from .utils import Displayer
from ...diary.model import value, optional_text, text
from ...sql import Interval

log = getLogger(__name__)


class DatabaseDisplayer(Displayer):

    @optional_text('Database')
    def _read_date(self, s, date):
        q = s.query(Interval.id).filter(Interval.dirty == True, Interval.permanent == False)
        total = q.count()
        today = q.filter(Interval.start <= date, Interval.finish > date).count()
        if not total:
            yield text('No dirty statistics')
        else:
            if today:
                yield value('Dirty statistics today', today)
            yield value('Total dirty statistics', total)
            yield text('Recalculate statistics')

    def _read_schedule(self, s, date, schedule):
        # todo - include?
        return
        yield



from collections import defaultdict
from logging import getLogger
from multiprocessing import cpu_count
from os.path import join, exists
from time import sleep

from psutil import NoSuchProcess

from ..commands.args import LOG, LOG_DIR
from ..common.date import now, format_seconds, time_to_local_time
from ..sql import PipelineType, Interval, Pipeline
from ..sql.tables.pipeline import sort_pipelines

log = getLogger(__name__)


def run_pipeline(config, type, *args, like=tuple(), worker=None, **extra_kargs):
    if type == PipelineType.PROCESS:
        run_process_pipeline(config, type, *args, like=like, worker=worker, **extra_kargs)
    else:
        from .pipeline import run_pipeline
        run_pipeline(config, type, like=like, worker=worker, **extra_kargs)


def run_process_pipeline(config, type, *args, like=tuple(), worker=None, **extra_kargs):
    if not worker:
        with config.db.session_context() as s:
            Interval.clean(s)
    with config.db.session_context(expire_on_commit=False) as s:
        pipelines = list(sort_pipelines(Pipeline.all(s, type, like=like, id=worker)))
    ProcessRunner(config, pipelines, *args, worker=worker, **extra_kargs).run()


def instantiate_pipeline(pipeline, config, *args, **kargs):
    kargs = dict(kargs)
    kargs.update(pipeline.kargs)  # this is where kargs from the config are added in
    log.debug(f'Instantiating {pipeline} with {args}, {kargs}')
    return pipeline.cls(config, *args, **kargs)


class ProcessRunner:

    def __init__(self, config, pipelines, *args, worker=None, n_cpu=cpu_count(), load=1, **kargs):
        if worker and len(pipelines) > 1: raise Exception('Worker with multiple pipelines')
        if not pipelines: raise Exception('No pipelines')
        self.__config = config
        self.__pipelines = pipelines
        self.__worker = worker
        self.__n_cpu = n_cpu
        self.__load = load
        self.__args = args
        self.__kargs = kargs
        self.__max_wait = 0
        self.__max_wait_procs = 0
        self.__max_wait_proc = None

    def run(self):
        if self.__worker or self.__n_cpu == 1:
            for pipeline in self.__pipelines:
                self.__run_local(pipeline)
        else:
            self.__run_commands(DependencyQueue(self.__config, self.__pipelines, self.__kargs))

    def __run_local(self, pipeline):
        log.info(f'Running pipeline {pipeline} locally with {self.__kargs}')
        instantiate_pipeline(pipeline, self.__config, *self.__args,
                             id=self.__worker, worker=bool(self.__worker), **self.__kargs).run()

    def __run_commands(self, queue):
        log.info('Scheduling worker pipelines')
        capacity = max(1, int(self.__n_cpu * self.__load))
        pipelines, popens = {}, []
        while True:
            try:
                pipeline, cmd, log_index = queue.pop()
                popen = self.__config.run_process(pipeline.cls, cmd, log_name(pipeline, log_index),
                                                  constraint=pipeline.id)
                pipelines[popen] = (pipeline, log_index)
                popens.append(popen)
                if len(popens) == capacity:
                    popens = self._run_til_next(pipelines, popens, queue)
            except EmptyException:
                if popens:
                    log.debug('Nothing new to add')
                    popens = self._run_til_next(pipelines, popens, queue)
                else:
                    log.debug('Done')
                    queue.shutdown()
                    log.info(f'Maximum wait {format_seconds(self.__max_wait)} for {self.__max_wait_proc} '
                             f'with {self.__max_wait_procs} processes')
                    return

    def _run_til_next(self, pipelines, popens, queue):
        queue.log()
        start = now()
        log.debug('Waiting for a subprocess to complete')
        while True:
            for i, popen in enumerate(popens):
                popen.poll()
                process = self.__config.get_process(pipelines[popen][0].cls, popen.pid)
                if popen.returncode is not None:
                    del popens[i]
                    duration = (now() - start).total_seconds()
                    log.debug(f'Waited {format_seconds(duration)}')
                    pipeline, log_index = pipelines.pop(popen)
                    if duration > self.__max_wait:
                        self.__max_wait = duration
                        self.__max_wait_procs = len(pipelines) + 1
                        self.__max_wait_proc = str(pipeline)
                    self.__config.delete_process(pipeline.cls, popen.pid)
                    queue.complete(pipeline, log_index)
                    if popen.returncode:
                        msg = f'Command "{popen.args}" exited with return code {popen.returncode} ' + \
                              f'see {process.log} for more info'
                        log.warning(msg)
                        self._copy_log(process.log)
                        self._abort(pipelines, popens)
                        raise Exception(msg)
                    else:
                        log.debug(f'Command "{fmt_cmd(popen.args)}" finished successfully')
                        return popens
            sleep(0.1)

    def _abort(self, pipelines, popens):
        for popen in popens:
            log.warning(f'Killing PID {popen.pid} ({popen.args})')
            try:
                popen.kill()
            except NoSuchProcess:
                pass
            pipeline, log_index = pipelines[popen]
            self.__config.delete_process(pipeline.cls, popen.pid)

    def _copy_log(self, filename, n_lines=20):
        dir = self.__config.args._format_path(LOG_DIR)
        path = join(dir, filename)
        if exists(path):
            with open(path) as input:
                lines = input.readlines()
                lines = lines[-n_lines:]
                log.debug(f'>>> Showing last {n_lines} from {path}')
                for line in lines:
                    line = line.strip()
                    log.debug(f'>>> {line}')
        else:
            log.warning(f'Cannot find {path}')


class EmptyException(Exception): pass


class DependencyQueue:

    def __init__(self, config, pipelines, kargs, min_missing=1, max_missing=20, gamma=0.4):
        self.__clean_pipelines(pipelines)
        self.__config = config
        self.__blocked = [pipeline for pipeline in pipelines if pipeline.blocked_by]
        self.__unblocked = [pipeline for pipeline in pipelines if not pipeline.blocked_by]
        self.__complete = []
        self.__active = {}  # pipeline: (instance, missing)
        self.__stats = {}  # pipeline: Stats
        self.__order = []
        self.__kargs = kargs
        self.__max_missing = max_missing
        self.__min_missing = min_missing
        self.__gamma = gamma
        self.__active_log_indices = defaultdict(lambda: set())
        self.__start = now()
        # clear out any junk from previous errors?
        for pipeline in self.__unblocked:
            self.__config.delete_all_processes(pipeline.cls, constraint=pipeline.id)

    def __clean_pipelines(self, pipelines):
        included = set(pipelines)
        for pipeline in pipelines:
            for i in reversed(range(len(pipeline.blocked_by))):
                blocker = pipeline.blocked_by[i]
                if blocker not in included:
                    log.warning(f'Removing {blocker} from blockers for {pipeline}')
                    del pipeline.blocked_by[i]
        log.debug('Cleaned pipelines')

    def complete(self, pipeline, log_index=None):
        # check if completed instance means that a pipeline is complete and, if so,
        # see if that unblocks others
        if log_index is not None:
            self.__stats[pipeline].finish(log_index)
            self.__active_log_indices[pipeline].remove(log_index)
        if self.__stats[pipeline]:
            log.info(f'{pipeline} complete ({self.__stats[pipeline].done})')
            self.__complete.append(pipeline)
            instance, missing = self.__active.pop(pipeline)
            if missing: raise Exception(f'Complete pipeline {pipeline} still has missing values {missing}')
            self.__order = [p for p in self.__order if p != pipeline]
            log.debug(f'Calling shutdown for {pipeline}')
            instance.shutdown()
            for i in reversed(range(len(self.__blocked))):
                if all(blocker in self.__complete for blocker in self.__blocked[i].blocked_by):
                    unblocked = self.__blocked.pop(i)
                    log.info(f'{pipeline} unblocks {unblocked}')
                    self.__unblocked.append(unblocked)
                    # self.__config.delete_all_processes(unblocked.cls)

    def pop(self):
        # unblocking takes some time, so do it step by step as we need more
        # add the new pipeline to the head of active
        while self.__unblocked:
            pipeline = self.__unblocked.pop()
            log.debug(f'Making {pipeline} active')
            instance = instantiate_pipeline(pipeline, self.__config, **self.__kargs)
            instance.startup()
            missing = instance.missing()
            self.__stats[pipeline] = Stats(pipeline, missing)
            if missing:
                log.debug(f'{pipeline}: {len(missing)} missing values')
                self.__active[pipeline] = instance, missing
                self.__order.insert(0, pipeline)
                break
            else:
                self.__active[pipeline] = instance, None
                self.__order.insert(0, pipeline)
                self.complete(pipeline)
                log.debug(f'{pipeline}: no missing data')
        for _ in range(len(self.__order)):  # at most, try each once
            pipeline = self.__order.pop(0)
            instance, missing = self.__active[pipeline]
            if missing:
                log_index = self.__unused_log_index(pipeline)
                missing_args, missing = self.__split_missing(pipeline, missing)
                cmd = instance.command_for_missing(pipeline, missing_args, log_name(pipeline, log_index))
                self.__active[pipeline] = instance, missing
                self.__order.append(pipeline)
                log.debug(f'{pipeline}: starting batch of {len(missing_args)} missing values')
                self.__stats[pipeline].start(log_index, len(missing_args))
                return pipeline, cmd, log_index
            else:
                log.debug(f'{pipeline} exhausted')
                self.__active[pipeline] = instance, None
                self.__order.append(pipeline)
        raise EmptyException()

    def __unused_log_index(self, pipeline):
        index = 0
        while index in self.__active_log_indices[pipeline]: index += 1
        self.__active_log_indices[pipeline].add(index)
        return index

    def log(self):
        log.info(f'Started {time_to_local_time(self.__start)}; '
                 f'duration {format_seconds((now() - self.__start).total_seconds())}; '
                 f'{len(self.__blocked)} blocked')
        for pipeline in self.__stats:
            log.info(str(self.__stats[pipeline]))

    def __log_efficiency(self):
        clock_time = (now() - self.__start).total_seconds()
        process_time = 0
        for pipeline in self.__stats:
            process_time += self.__stats[pipeline].duration_individual
        speedup = process_time / clock_time
        log.info(f'Clock time: {format_seconds(clock_time)}; Process time: {format_seconds(process_time)}; '
                 f'Speedup: x{speedup:.1f}')
        log.info(f'Missing args: min {self.__min_missing}; max {self.__max_missing}; gamma {self.__gamma}')

    def __split_missing(self, pipeline, missing):
        # this (min, min) is a bit weird but makes sense, i think
        n = min(len(missing),
                max(self.__min_missing,
                    min(self.__max_missing, int(pow(self.__stats[pipeline].total, self.__gamma)))))
        return missing[:n], missing[n:]

    def shutdown(self):
        self.log()
        self.__log_efficiency()
        if self.__blocked:
            log.warning(f'{len(self.__blocked)} pipelines still blocked')
            for pipeline in self.__blocked:
                log.warning(f'{pipeline}: blocked by '
                            f'{", ".join(str(b) for b in pipeline.blocked_by if b not in self.__complete)}')
        else:
            log.debug('All pipelines unblocked')


class Stats:

    def __init__(self, pipeline, missing):
        self.__pipeline = pipeline
        self.active = 0
        self.total = len(missing) if missing else 0
        self.done = 0
        self.duration_overall = 0
        self.duration_individual = 0
        self.__start_overall = now()
        self.__start_individual = {}
        self.__size = {}

    def start(self, index, n):
        self.__start_individual[index] = now()
        self.__size[index] = n
        self.active += 1

    def finish(self, index):
        log.info(f'{self.__pipeline}: {self.__size[index]} completed')
        self.active -= 1
        self.done += self.__size[index]
        self.duration_individual += (now() - self.__start_individual[index]).total_seconds()
        if self:
            self.duration_overall = (now() - self.__start_overall).total_seconds()

    def __bar(self, width):
        solid = int(width * self.done / self.total) if self.total else width
        blank = width - solid
        bar = '-' * blank + '#' * solid
        if self:
            overall_secs = format_seconds(self.duration_overall)
            individual_secs = format_seconds(self.duration_individual)
            secs = f' {overall_secs} / {individual_secs}'
            bar = bar[:-len(secs)] + secs
        return bar

    def __label(self):
        label = str(self.__pipeline)
        for suffix in ('Reader', 'Calculator'):
            if label.endswith(suffix):
                label = label[:-len(suffix)+1] + '%'
        return label

    def __str__(self):
        return f'{self.__label():>13s} {self.__pipeline.id:<2d} {self.active:2d} {self.done:4d}/{self.total:<4d} {self.__bar(40)}'

    def __bool__(self):
        return self.done == self.total


def log_name(pipeline, log_index):
    return f'{pipeline}.{log_index}.{LOG}'


def fmt_cmd(cmd, max=400):
    if len(cmd) > max:
        n1 = int(0.8 * max)
        n2 = len(cmd) - (max - n1 - 3)
        return cmd[:n1] + '...' + cmd[n2:]
    else:
        return cmd




import datetime as dt
from contextlib import contextmanager
from logging import getLogger, WARNING, INFO, DEBUG
from shutil import rmtree
from tempfile import mkdtemp
from unittest import TestCase

from ch2 import PROGNAME
from ch2.common.args import mm
from ch2.common.io import data_hash
from ch2.common.log import configure_log
from ch2.commands.args import make_parser, NamespaceWithVariables, DB_VERSION
from ch2.common.names import USER
from ch2.common.user import make_user_database
from ch2.sql.config import Config
from ch2.sql.support import Base

log = getLogger(__name__)


class LogTestCase(TestCase):

    def setUp(self):
        configure_log(PROGNAME, '/tmp/ch2-test.log', verbosity=5, levels={
            'sqlalchemy': WARNING,
            'matplotlib': INFO,
            'bokeh': DEBUG,
            'tornado': INFO,
            'sentinelsat': DEBUG,
            'werkzeug': DEBUG,
            'ch2': DEBUG,
            '__main__': DEBUG,
            'tests': DEBUG
        })


def random_test_user(args=(mm(USER), 'postgres')):
    parser = make_parser()
    ns = NamespaceWithVariables._from_ns(parser.parse_args(args=args), PROGNAME, DB_VERSION)
    config = Config(ns)
    user = data_hash(str(dt.datetime.now()))[:6]
    log.info(f'User/database {user}')
    user_config = make_user_database(config, user, '')
    log.info('Creating tables')
    Base.metadata.create_all(user_config.db.engine)
    return user


@contextmanager
def TempDirOnSuccess():
    dir = mkdtemp()
    try:
        yield dir
    except:
        log.warning(f'Leaving {dir} for inspection')
        dir = None
        raise
    finally:
        if dir: rmtree(dir)



from logging import getLogger
from os.path import splitext, basename

from pygeotile.point import Point

from .utils import AbortImportButMarkScanned, ProcessFitReader
from ..pipeline import LoaderMixin
from ...commands.args import DEFAULT
from ...commands.upload import ACTIVITY
from ...common.date import to_time
from ...diary.model import TYPE, EDIT
from ...fit.format.records import fix_degrees, merge_duplicates, no_bad_values
from ...fit.profile.profile import read_fit
from ...lib.io import split_fit_path
from ...names import N, T, U, Sports, S
from ...sql.database import StatisticJournalText
from ...sql.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...sql.tables.statistic import StatisticJournalFloat, STATISTIC_JOURNAL_CLASSES, StatisticName, \
    StatisticJournalType, StatisticJournal, STATISTIC_JOURNAL_TYPES
from ...sql.tables.topic import ActivityTopicField, ActivityTopic, ActivityTopicJournal
from ...sql.utils import add
from ...srtm.bilinear import bilinear_elevation_from_constant

log = getLogger(__name__)


# duplicate data in
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24715592701_tap-sync-18690-cc1dd93225119215a1ea87c584a974ce.fit
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24718989709_tap-sync-18690-effaaaffdd06b9419991471bd92d53d5.fit

class ActivityReader(LoaderMixin, ProcessFitReader):

    KIT = 'kit'

    def __init__(self, *args, sport_to_activity=None, record_to_db=None, **kargs):
        self.sport_to_activity = self._assert('sport_to_activity', sport_to_activity)
        self.record_to_db = [(field, title, units, STATISTIC_JOURNAL_CLASSES[type])
                             for field, (title, units, type)
                             in self._assert('record_to_db', record_to_db).items()]
        self.add_elevation = not any(title == T.ELEVATION for (field, title, units, type) in self.record_to_db)
        self.__ajournal = None  # save for coverage
        super().__init__(*args, sub_dir=ACTIVITY, **kargs)

    def _startup(self, s):
        self.__srtm1 = bilinear_elevation_from_constant(s)
        super()._startup(s)
        for field, title, units, cls in self.record_to_db:
            self._provides(s, title, STATISTIC_JOURNAL_TYPES[cls], units, None,
                           f'The value of field {field} in the FIT record.')
        self._provides(s, T.ELAPSED_TIME, StatisticJournalType.FLOAT, U.S, None,
                       'The time since the start of the activity')
        self._provides(s, T.SPHERICAL_MERCATOR_X, StatisticJournalType.FLOAT, U.M, None,
                       'The Web Mercator EPSG:3857 X coordinate')
        self._provides(s, T.SPHERICAL_MERCATOR_Y, StatisticJournalType.FLOAT, U.M, None,
                       'The Web Mercator EPSG:3857 Y coordinate')
        self._provides(s, T.SRTM1_ELEVATION, StatisticJournalType.FLOAT, U.M, None,
                       'The elevation from SRTM1 at the GPS location')
        self._provides(s, T.KIT, StatisticJournalType.TEXT, None, None,
                       'The kit used in the activity')
        self._provides(s, T.DEVICE, StatisticJournalType.TEXT, None, None,
                       'The device used to record the activity')
        # also coverages - see _read

    @staticmethod
    def _read_kit(path):
        _, kit = split_fit_path(path)
        if kit:
            return kit
        else:
            log.debug(f'No {N.KIT} in {path}')
            return None

    @staticmethod
    def _read_device(records):
        for record in records:
            if record.name == 'device_info' and 'garmin_product' in record.data:
                return record.value.garmin_product
        return None

    def _read_data(self, s, file_scan):
        log.info('Reading activity data from %s' % file_scan)
        records = self.parse_records(read_fit(file_scan.path))
        kit = self._read_kit(file_scan.path)
        device = self._read_device(records)
        ajournal, activity_group, first_timestamp = self._create_activity(s, file_scan, kit, records)
        return ajournal, (ajournal, activity_group, first_timestamp, file_scan, kit, device, records)

    @staticmethod
    def parse_records(data):
        log.debug('Parsing records')
        records = ActivityReader.read_fit_file(data, merge_duplicates, fix_degrees, no_bad_values)
        log.debug('Parsed')
        return records

    @staticmethod
    def read_sport(path, records):
        try:
            return ActivityReader._first(path, records, 'sport').value.sport.lower()
        except AbortImportButMarkScanned:
            # alternative for some garmin devices (florian)
            return ActivityReader._first(path, records, 'session').value.sport.lower()

    @staticmethod
    def read_first_timestamp(path, records):
        return ActivityReader._first(path, records, 'event', 'record').value.timestamp

    @staticmethod
    def read_last_timestamp(path, records):
        return ActivityReader._last(path, records, 'event', 'record').value.timestamp

    def _create_activity(self, s, file_scan, kit, records):
        first_timestamp = self.read_first_timestamp(file_scan.path, records)
        last_timestamp = self.read_last_timestamp(file_scan.path, records)
        log.debug(f'Time range: {first_timestamp.timestamp()} - {last_timestamp.timestamp()}')
        sport = self.read_sport(file_scan.path, records)
        activity_group = self._activity_group(s, file_scan.path, sport, self.sport_to_activity, kit)
        log.info(f'{activity_group} from {sport} / {kit}')
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          file_hash_id=file_scan.file_hash_id))
        return ajournal, activity_group, first_timestamp

    def _activity_group(self, s, path, sport, lookup, kit):
        log.debug(f'Current lookup: {lookup}; sport: {sport}; kit: {kit}')
        if isinstance(lookup, str):
            return self._lookup_activity_group(s, lookup)
        if sport in lookup:
            return self._activity_group(s, path, sport, lookup[sport], kit)
        if N.KIT in lookup and kit in lookup[N.KIT]:
            return self._activity_group(s, path, sport, lookup[N.KIT][kit], kit)
        if DEFAULT in lookup:
            return self._activity_group(s, path, sport, lookup[DEFAULT], kit)
        log.warning('Unrecognised sport: "%s" in %s' % (sport, path))
        if sport in (Sports.SPORT_GENERIC,):
            raise AbortImportButMarkScanned(f'Ignoring {sport} entry')
        else:
            raise Exception(f'There is no group configured for {sport} entries in the FIT file.')

    def _lookup_activity_group(self, s, name):
        activity_group = ActivityGroup.from_name(s, name)
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                log.info('Available activity group:')
                for activity_group in activities:
                    log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    @staticmethod
    def _save_name(s, ajournal, file_scan):
        from ...config.database import add_activity_topic_field
        log.debug('Saving name')
        # first, do we have the 'Name' field defined?
        # this should be triggered at most once per group if it was not already defined
        root = s.query(ActivityTopic). \
            filter(ActivityTopic.title == ActivityTopic.ROOT,
                   ActivityTopic.activity_group == ajournal.activity_group).one_or_none()
        if not root:
            root = add(s, ActivityTopic(title=ActivityTopic.ROOT, description=ActivityTopic.ROOT_DESCRIPTION,
                                        activity_group=ajournal.activity_group))
        if not s.query(ActivityTopicField). \
                join(StatisticName). \
                filter(StatisticName.name == N.NAME,
                       StatisticName.owner == ActivityTopic,
                       ActivityTopicField.activity_topic == root).one_or_none():
           add_activity_topic_field(s, root, T.NAME, -10, StatisticJournalType.TEXT,
                                     ajournal.activity_group, model={TYPE: EDIT},
                                     description=ActivityTopicField.NAME_DESCRIPTION)
        # second, do we already have a journal for this file, or do we need to add one?
        source = s.query(ActivityTopicJournal). \
            filter(ActivityTopicJournal.file_hash_id == file_scan.file_hash_id). \
            one_or_none()
        if not source:
            source = add(s, ActivityTopicJournal(file_hash_id=file_scan.file_hash_id,
                                                 activity_group=ajournal.activity_group))
        # and third, does that journal contain a name?
        if not s.query(StatisticJournal). \
                join(StatisticName). \
                filter(StatisticJournal.source == source,
                       StatisticName.owner == ActivityTopic,
                       StatisticName.name == N.NAME).one_or_none():
            value = splitext(basename(file_scan.path))[0]
            StatisticJournalText.add(s, N.NAME, None, None, ActivityTopic,
                                     source, value, ajournal.start)

    def _load_data(self, s, loader, data):

        ajournal, activity_group, first_timestamp, file_scan, kit, device, records = data
        timespan, warned, logged, last_timestamp = None, 0, 0, to_time(0.0)

        log.debug(f'Loading {self.record_to_db}')

        def is_event(record, *types):
            if record.name == 'event' and record.value.event == 'timer' and record.value.event_type in types:
                log.debug(f'{types} at {record.timestamp}')
                return True
            else:
                return False

        have_timespan = any(is_event(record, 'start') for record in records)
        only_records = list(filter(lambda x: x.name == 'record', records))
        final_timestamp = only_records[-1].timestamp

        if kit: loader.add_data(N.KIT, ajournal, kit, ajournal.start)
        if device: loader.add_data(N.DEVICE, ajournal, device, ajournal.start)

        self._save_name(s, ajournal, file_scan)
        self.__ajournal = ajournal

        if not have_timespan:
            first_timestamp = only_records[0].timestamp
            log.warning('Experimental handling of data without timespans')
            timespan = add(s, ActivityTimespan(activity_journal=ajournal, start=first_timestamp, finish=final_timestamp))

        for record in records:

            if have_timespan and is_event(record, 'start'):
                if timespan:
                    log.warning('Ignoring start with no corresponding stop (possible lost data?)')
                else:
                    timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                       start=record.value.timestamp,
                                                       finish=record.value.timestamp))

            elif record.name == 'record':
                if record.value.timestamp > last_timestamp:
                    lat, lon, timestamp = None, None, record.value.timestamp
                    # elapsed time is not customizable because it needs extra processing
                    loader.add_data(T.ELAPSED_TIME, ajournal,
                                    (record.value.timestamp - first_timestamp).total_seconds(), timestamp)
                    # customizable loader
                    for field, title, units, type in self.record_to_db:
                        value = record.data.get(field, None)
                        if logged < 3:
                            log.debug(f'{title} = {value}')
                        if value is not None:
                            value = value[0][0]
                            if units == U.KM:  # internally everything uses M
                                value /= 1000
                            loader.add_data(title, ajournal, value, timestamp)
                            if title == T.LATITUDE:
                                lat = value
                            elif title == T.LONGITUDE:
                                lon = value
                    logged += 1
                    # values derived from lat/lon
                    if lat is not None and lon is not None:
                        x, y = Point.from_latitude_longitude(lat, lon).meters
                        loader.add_data(N.SPHERICAL_MERCATOR_X, ajournal, x, timestamp)
                        loader.add_data(N.SPHERICAL_MERCATOR_Y, ajournal, y, timestamp)
                        if self.add_elevation:
                            elevation = self.__srtm1.elevation(lat, lon)
                            if elevation:
                                loader.add_data(N.SRTM1_ELEVATION, ajournal, elevation, timestamp)
                else:
                    log.warning('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                (file_scan.path, record.value.timestamp))
                last_timestamp = record.value.timestamp
                if not have_timespan:
                    ajournal.finish = record.value.timestamp

            elif have_timespan and is_event(record, 'stop_all', 'stop'):
                if timespan:
                    timespan.finish = record.value.timestamp
                    ajournal.finish = record.value.timestamp
                    timespan = None
                else:
                    log.debug('Ignoring stop with no corresponding start (possible lost data?)')

        if timespan:
            log.warning('Cleaning up dangling timespan')
            timespan.finish = final_timestamp

    def _read(self, s, path):
        loader = super()._read(s, path)
        for title, percent in loader.coverage_percentages():
            StatisticJournalFloat.add(s, T._cov(title), U.PC, S.join(S.MIN, S.AVG), self.owner_out,
                                      self.__ajournal, percent, self.__ajournal.start,
                                      description=f'Coverage (% of FIT records with data) for {title}.')




import datetime as dt

from bokeh.io import output_file
from bokeh.layouts import gridplot
from bokeh.plotting import show

from ch2.data import *
from ch2.jupyter.decorator import template
from ch2.names import Names as N, like


@template
def health():

    '''
    # Health
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')
    health = std_health_statistics(s)

    '''
    ## Health and Fitness
    '''

    output_file(filename='/dev/null')

    fitness, fatigue = like(N.FITNESS_ANY, health.columns), like(N.FATIGUE_ANY, health.columns)
    colours = ['black'] * len(fitness) + ['red'] * len(fatigue)
    alphas = [1.0] * len(fitness) + [0.5] * len(fatigue)
    ff = multi_line_plot(900, 300, N.TIME, fitness + fatigue, health, colours, alphas=alphas)
    xrange = ff.x_range if ff else None
    add_multi_line_at_index(ff, N.TIME, fitness + fatigue, health, colours, alphas=alphas, index=-1)
    atd = std_distance_time_plot(900, 200, health, x_range=xrange)
    show(gridplot([[ff], [atd]]))



from json import loads
from logging import getLogger

from sqlalchemy import func, text

from .power import PowerModel
from .utils import ActivityGroupProcessCalculator, ProcessCalculator
from ..pipeline import LoaderMixin
from ...common.date import local_time_to_time
from ...data.sector import find_and_add_sector_journals, add_sector_statistics
from ...names import simple_name
from ...sql import Timestamp, Constant, ActivityJournal
from ...sql.tables.sector import SectorGroup, Sector, SectorClimb, SectorJournal

log = getLogger(__name__)


class SectorCalculator(LoaderMixin, ActivityGroupProcessCalculator):
    '''
    Run through new activity journals, find any sectors that match, populate the SectorJournal,
    and add associated statistics.
    '''

    def __init__(self, *args, power_model=None, **kargs):
        super().__init__(*args, **kargs)
        self.__power_model_ref = power_model

    def _startup(self, s):
        super()._startup(s)
        Sector.provides(s, self)
        SectorClimb.provides(s, self)
        SectorJournal.clean(s)
        if self.__power_model_ref:
            self.__power_model = PowerModel(**loads(Constant.from_name(s, self.__power_model_ref).at(s).value))
        else:
            log.warning('No power model configured for sectors')
            self.__power_model = None

    def _run_one(self, missed):
        start = local_time_to_time(missed)
        with self._config.db.session_context() as s:
            ajournal = self._get_source(s, start)
            if self.__power_model:
                power_model = self.__power_model.expand(s, ajournal.start, default_owner=Constant,
                                                        default_activity_group=self.activity_group)
            else:
                power_model = None
            log.debug(f'Power: {self.__power_model_ref}: {power_model}')
            with Timestamp(owner=self.owner_out, source=ajournal).on_success(s):
                self._run_activity_journal(s, ajournal, power_model)

    def _run_activity_journal(self, s, ajournal, power_model):
        if ajournal.route_edt:
            for sector_group in s.query(SectorGroup). \
                    filter(func.ST_Distance(SectorGroup.centre, ajournal.centre) < SectorGroup.radius). \
                    all():
                count = 0
                for sjournal in find_and_add_sector_journals(s, sector_group, ajournal):
                    loader = self._get_loader(s, add_serial=False)
                    add_sector_statistics(s, sjournal, loader, power_model=power_model)
                    loader.load()
                    count += 1
                if count:
                    log.info(f'Found {count} sectors for activity on {ajournal.start}')


class NewSectorCalculator(LoaderMixin, ProcessCalculator):
    '''
    Called when a new sector is added via the GUI.
    '''

    def __init__(self, *args, activity_group=None, new_sector_id=None, **kargs):
        super().__init__(*args, **kargs)
        self.activity_group = self._assert('activity_group', activity_group)
        self.new_sector_id = int(self._assert('new_sector_id', new_sector_id))

    def _startup(self, s):
        super()._startup(s)
        SectorJournal.clean(s)
        self.sector = s.query(Sector).filter(Sector.id == int(self.new_sector_id)).one()
        self.sector.sector_group  # eager load
        # we don't provide anything because they are statistics identical to those already declared by
        # SectorCalculator

    def _missing(self, s):
        q = text('''
select aj.id
  from activity_journal as aj,
       source as sc,
       activity_group as ag,
       sector as s,
       sector_group as sg
 where s.id = :new_sector_id
   and s.sector_group_id = sg.id
   and st_distance(sg.centre, aj.centre) < sg.radius
   and sc.id = aj.id
   and ag.id = sc.activity_group_id
   and ag.name = :activity_group
   and aj.route_edt is not null
        ''')
        log.debug(q)
        log.debug(f'new_sector_id: {self.new_sector_id}; activity_group: {self.activity_group}')
        return [str(row[0]) for row in
                s.connection().execute(q, new_sector_id=self.new_sector_id,
                                       activity_group=simple_name(self.activity_group)).fetchall()]

    def _run_one(self, missed):
        with self._config.db.session_context() as s:
            count = 0
            ajournal = s.query(ActivityJournal).filter(ActivityJournal.id == int(missed)).one()
            for sjournal in find_and_add_sector_journals(s, self.sector.sector_group, ajournal,
                                                         sector_id=self.new_sector_id):
                # fake owner - we're patching in the above for new sectors
                loader = self._get_loader(s, add_serial=False, owner=SectorCalculator)
                # no need for power_model - called for user-defined sectors, not climbs
                add_sector_statistics(s, sjournal, loader)
                loader.load()
                count += 1
            if count:
                log.info(f'Found {count} sectors for activity on {ajournal.start}')



from enum import IntEnum
from logging import getLogger

from geoalchemy2 import Geometry, WKTElement, Geography
from geoalchemy2.shape import to_shape
from sqlalchemy import Column, Integer, Text, ForeignKey, Float, UniqueConstraint, or_, func, text
from sqlalchemy.orm import relationship

from .source import GroupedSource, Source, SourceType
from .statistic import StatisticJournalType
from ..support import Base
from ..triggers import add_child_ddl, add_text
from ..types import ShortCls, UTC, point
from ..utils import add, SPHM_SRID, WGS84_SRID
from ...common.geo import utm_srid
from ...common.plot import CYAN
from ...names import N, T, U, S

log = getLogger(__name__)

DEFAULT_GROUP_RADIUS_KM = 200


@add_child_ddl(Source)
class SectorJournal(GroupedSource):

    __tablename__ = 'sector_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    # do not use on delete cascade here since it leaves source entries without children
    # instead, to guarantee consistency, call clean()
    sector_id = Column(Integer, ForeignKey('sector.id', ondelete='set null'), index=True)
    sector = relationship('Sector')
    # do not use on delete cascade here since it leaves source entries without children
    # instead, to guarantee consistency, call clean()
    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='set null'),
                                 index=True)
    activity_journal = relationship('ActivityJournal', foreign_keys=[activity_journal_id])
    start_fraction = Column(Float, nullable=False)
    finish_fraction = Column(Float, nullable=False)
    # these duplicate data (can be extracted from st_linesubstring and the activity route)
    # but simplify time_range below.
    start_time = Column(UTC, nullable=False)
    finish_time = Column(UTC, nullable=False)
    # this duplicates data because it's useful and a pain to calculate separately
    start_distance = Column(Float, nullable=False)
    finish_distance = Column(Float, nullable=False)
    start_elevation = Column(Float, nullable=False)
    finish_elevation = Column(Float, nullable=False)
    UniqueConstraint(sector_id, activity_journal_id, start_fraction, finish_fraction)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SECTOR
    }

    def time_range(self, s):
        return self.start_time, self.finish_time

    @classmethod
    def clean(cls, s):
        q1 = s.query(SectorJournal.id). \
            filter(or_(SectorJournal.sector_id == None,
                       SectorJournal.activity_journal_id == None)).cte()
        s.query(Source).filter(Source.id.in_(q1)).delete(synchronize_session=False)


class SectorGroup(Base):
    '''
    this is not owned by anything (sectors are owned individually).
    rather, it defines a place where we are interested in all kinds of sectors -
    typically, the geographical region where the user lives.
    we could even try setting it algorithmically from a few activities.
    '''

    __tablename__ = 'sector_group'

    id = Column(Integer, primary_key=True)
    srid = Column(Integer, nullable=False)
    centre = Column(Geography('Point', srid=WGS84_SRID), nullable=False)
    radius = Column(Float, nullable=False)  # metres
    title = Column(Text, nullable=False, unique=True)
    UniqueConstraint(centre, radius)

    @classmethod
    def add(cls, s, centre, radius_km, title, delete=False):
        '''
        if delete is True, nearby similar entries are deleted before insertion.
        if delete is False, and there is already a nearby similar entry, it is returned.
        otherwise, insertion is attempted without deletion.
        centre is (lon, lat)
        '''
        lon, lat = centre
        srid = utm_srid(lat, lon)
        centre = text(point(lon, lat))
        radius = radius_km * 1000
        if delete is not None:
            query = s.query(SectorGroup). \
                filter(func.ST_Distance(SectorGroup.centre, centre) < radius,
                       SectorGroup.srid == srid)
            n = query.count()
            if delete:
                if n:
                    log.warning(f'Deleting {n} previous sector groups')
                    query.delete(synchronize_session=False)
            else:
                if n:
                    sector_group = query.first()
                    log.info(f'Using previously defined sector group "{sector_group.title}"')
                    return sector_group
        return add(s, SectorGroup(srid=srid, centre=centre, radius=radius, title=title))


class SectorType(IntEnum):
    '''
    different owners don't need to define their own types.
    the subclasses are mainly for display (climb has elevation and category and is displayed differently)
    '''

    SECTOR = 0
    CLIMB = 1


class Sector(Base):

    __tablename__ = 'sector'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast delete of subtypes
    sector_group_id = Column(Integer, ForeignKey('sector_group.id', ondelete='cascade'), nullable=False)
    sector_group = relationship('SectorGroup')
    route = Column(Geometry('LineString'), nullable=False)
    title = Column(Text)  # null will have name generated on display
    owner = Column(ShortCls, nullable=False, index=True)
    # null because set later (calculated from route but stored for efficiency)
    distance = Column(Float)
    start = Column(Geometry('LineString'))
    finish = Column(Geometry('LineString'))
    hull = Column(Geometry('Polygon'))

    __mapper_args__ = {
        'polymorphic_identity': SectorType.SECTOR,
        'polymorphic_on': 'type'
    }

    @classmethod
    def provides(cls, s, pipeline):
        pipeline._provides(s, T.SECTOR_TIME, StatisticJournalType.FLOAT, U.S, S.join(S.MIN, S.CNT, S.MSR),
                           'The time to complete the sector.')
        pipeline._provides(s, T.SECTOR_DISTANCE, StatisticJournalType.FLOAT, U.KM, None,
                           'The sector distance.')

    def add_statistics(self, s, sjournal, loader, **kargs):
        loader.add_data(N.SECTOR_TIME, sjournal, (sjournal.finish_time - sjournal.start_time).total_seconds(),
                        sjournal.start_time)
        loader.add_data(N.SECTOR_DISTANCE, sjournal, sjournal.finish_distance - sjournal.start_distance,
                        sjournal.start_time)

    def read_path(self, s):
        sql = text(f'''
select st_astext(st_transform(st_setsrid(s.route, sg.srid), {SPHM_SRID}))
  from sector as s,
       sector_group as sg
 where s.id = :sector_id
   and sg.id = s.sector_group_id
    ''')
        row = s.connection().execute(sql, sector_id=self.id).fetchone()
        return to_shape(WKTElement(row[0])).xy

    def display(self, s, fx, fy, ax, cm=1.5):
        xs, ys = self.read_path(s)
        ax.plot([fx(x) for x in xs], [fy(y) for y in ys], color=CYAN)


@add_child_ddl(Sector)
class SectorClimb(Sector):

    __tablename__ = 'sector_climb'

    id = Column(Integer, ForeignKey('sector.id', ondelete='cascade'), primary_key=True)
    category = Column(Text)
    elevation = Column(Float, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SectorType.CLIMB
    }

    # these don't delegate to parent because we want to use different names for basically the same thing

    @classmethod
    def provides(cls, s, pipeline):
        pipeline._provides(s, T.CLIMB_ELEVATION, StatisticJournalType.FLOAT, U.M, S.join(S.MAX, S.SUM, S.MSR),
                           'The difference in elevation between start and end of the climb.')
        pipeline._provides(s, T.CLIMB_DISTANCE, StatisticJournalType.FLOAT, U.KM, S.join(S.MAX, S.SUM, S.MSR),
                           'The distance travelled during the climb.')
        pipeline._provides(s, T.CLIMB_TIME, StatisticJournalType.FLOAT, U.S, S.join(S.MAX, S.SUM, S.MSR),
                           'The time spent on the climb.')
        pipeline._provides(s, T.CLIMB_GRADIENT, StatisticJournalType.FLOAT, U.PC,  S.join(S.MAX, S.MSR),
                           'The average inclination of the climb (elevation / distance).')
        pipeline._provides(s, T.VERTICAL_POWER, StatisticJournalType.FLOAT, U.W,  S.join(S.MAX, S.MSR),
                           'The estimated power for the climb considering only weight / height.')
        pipeline._provides(s, T.CLIMB_CATEGORY, StatisticJournalType.TEXT, None, None,
                           'The climb category (text, "4" to "1" and "HC").')

    def add_statistics(self, s, sjournal, loader, power_model=None, **kargs):
        # these are taken from the actial activity, not the sector, so will vary from match to match
        elevation = sjournal.finish_elevation - sjournal.start_elevation
        distance = sjournal.finish_distance - sjournal.start_distance
        loader.add_data(N.CLIMB_ELEVATION, sjournal, elevation, sjournal.start_time)
        loader.add_data(N.CLIMB_DISTANCE, sjournal, distance, sjournal.start_time)
        loader.add_data(N.CLIMB_TIME, sjournal, (sjournal.finish_time - sjournal.start_time).total_seconds(),
                        sjournal.start_time)
        loader.add_data(N.CLIMB_GRADIENT, sjournal, elevation / (10 * distance), sjournal.start_time)
        if self.category:
            loader.add_data(N.CLIMB_CATEGORY, sjournal, self.category, sjournal.start_time)
        if power_model:
            power = self._estimate_power(s, sjournal, power_model)
            loader.add_data(N.VERTICAL_POWER, sjournal, power, sjournal.start_time)

    def _estimate_power(self, s, sjournal, power_model, g=9.8):
        from ...data.query import interpolate
        from ...pipeline.read.activity import ActivityReader
        speed_start = interpolate(s, sjournal.activity_journal, N.SPEED, ActivityReader, sjournal.start_time)
        speed_finish = interpolate(s, sjournal.activity_journal, N.SPEED, ActivityReader, sjournal.finish_time)
        total_weight = power_model.bike_model.bike_weight + power_model.rider_weight
        energy_after = 0.5 * total_weight * speed_finish ** 2 + total_weight * g * self.elevation
        energy_before = 0.5 * total_weight * speed_start ** 2
        return (energy_after - energy_before) / (sjournal.finish_time - sjournal.start_time).total_seconds()

    def read_centroid(self, s):
        sql = text('''
      with point as (select st_centroid(st_transform(st_setsrid(s.route, sg.srid), 3785)) as point
                       from sector as s,
                            sector_group as sg
                      where s.sector_group_id = sg.id
                        and s.id = :sector_id)
    select st_x(point), st_y(point)
      from point; 
    ''')
        row = s.connection().execute(sql, sector_id=self.id).fetchone()
        return row[0], row[1]

    def display(self, s, fx, fy, ax, cm=1.5):
        x, y = self.read_centroid(s)
        ax.plot(fx(x), fy(y), marker='^', color=CYAN, markersize=cm*3)




from bokeh.plotting import output_file, show

from ch2.data import *
from ch2.jupyter.decorator import template
from ch2.names import N
from ch2.pipeline.owners import *


@template
def similar_activities(local_time):

    f'''
    # Similar Activities: {local_time.split()[0]}
    '''

    '''
    $contents
    '''

    '''
    ## Build Maps
    
    Loop over activities, retrieve data, and construct maps. 
    '''

    s = session('-v2')

    maps = [map_thumbnail(100, 120, data)
            for data in (Statistics(s, activity_journal=similar[0]).
                             by_name(ActivityReader, N.SPHERICAL_MERCATOR_X, N.SPHERICAL_MERCATOR_Y).
                             by_name(ActivityCalculator, N.ACTIVE_DISTANCE, N.ACTIVE_TIME).df
                         for similar in nearby_activities(s, local_time=local_time))
            if len(data[N.SPHERICAL_MERCATOR_X].dropna()) > 10]

    print(f'Found {len(maps)} activities')

    '''
    ## Display Maps
    '''

    output_file(filename='/dev/null')
    show(htile(maps, 8))




from logging import getLogger

from werkzeug import Request, run_simple
from werkzeug.exceptions import HTTPException, BadRequest
from werkzeug.routing import Map, Rule
from werkzeug.wrappers.json import JSONMixin

from .json import JsonResponse
from .middleware import CsrfCheck
from .servlets.configure import Configure
from .servlets.diary import Diary
from .servlets.image import Thumbnail, Sparkline
from .servlets.jupyter import Jupyter
from .servlets.kit import Kit
from .servlets.route import Route
from .servlets.search import Search
from .servlets.sector import Sector
from .servlets.statistics import Statistics
from .servlets.upload import Upload
from .static import Static
from ..commands.args import LOG, WEB, SERVICE, VERBOSITY, BIND, PORT, WARN, SECURE, IMAGE_DIR, \
    NOTEBOOK_DIR, JUPYTER
from ..common.args import mm
from ..common.names import BASE
from ..lib.server import BaseController
from ..sql import SystemConstant

log = getLogger(__name__)


MAX_MSG = 1000

DATA = 'data'
REDIRECT = 'redirect'

GET = 'GET'
PUT = 'PUT'
POST = 'POST'


class JSONRequest(Request, JSONMixin):
    pass


class WebController(BaseController):

    def __init__(self, config, max_retries=1, retry_secs=1):
        super().__init__(WEB, config, WebServer, max_retries=max_retries, retry_secs=retry_secs)
        args = config.args
        self.__warn_data = args[WARN + '-' + DATA]
        self.__warn_secure = args[WARN + '-' + SECURE]
        self.__notebook_dir = args[NOTEBOOK_DIR]
        self.__thumbnail_dir = args[IMAGE_DIR]
        self.__jupyter = args[JUPYTER]

    def _build_cmd_and_log(self, ch2):
        log_name = 'web-service.log'
        cmd = f'{ch2} {mm(VERBOSITY)} 0 {mm(LOG)} {log_name} {mm(BASE)} {self._config.args[BASE]} ' \
              f'{WEB} {SERVICE} {mm(WEB + "-" + BIND)} {self._bind} {mm(WEB + "-" + PORT)} {self._port} ' \
              f'{mm(JUPYTER)} {self.__jupyter} ' \
              f'{mm(IMAGE_DIR)} {self.__thumbnail_dir} {mm(NOTEBOOK_DIR)} {self.__notebook_dir}'
        if self.__warn_data: cmd += f' {mm(WARN + "-" + DATA)}'
        if self.__warn_secure: cmd += f' {mm(WARN + "-" + SECURE)}'
        return cmd, log_name

    def _run(self):
        log.debug(f'Binding to {self._bind}:{self._port}')
        run_simple(self._bind, self._port,
                   CsrfCheck(WebServer(self._config, warn_data=self.__warn_data, warn_secure=self.__warn_secure)),
                   use_debugger=self._dev, use_reloader=self._dev)

    def _cleanup(self):
        # default in case starting without a database
        self._config.delete_constant(SystemConstant.WEB_URL, default=None)

    def _status(self, running):
        if running:
            print(f'    {self.connection_url()}')
        print()

    def connection_url(self):
        return self._config.get_constant(SystemConstant.WEB_URL, none=True)


def error(exception):
    def handler(*args, **kwargs):
        raise exception()
    return handler


class WebServer:

    def __init__(self, config, warn_data=False, warn_secure=False):
        self.__config = config
        self.__warn_data = warn_data
        self.__warn_secure = warn_secure

        configure = Configure(config)
        diary = Diary()
        jupyter = Jupyter(config)
        kit = Kit()
        route = Route()
        sector = Sector(config)
        static = Static('.static')
        statistics = Statistics()
        upload = Upload(config)
        thumbnail = Thumbnail(config)
        sparkline = Sparkline(config)
        search = Search()

        self.url_map = Map([

            Rule('/api/configure/profiles', endpoint=self.check(configure.read_profiles, config=False), methods=(GET,)),
            Rule('/api/configure/initial', endpoint=self.check(configure.write_profile, config=False), methods=(POST,)),
            Rule('/api/configure/delete', endpoint=self.check(configure.delete, config=False, busy=True), methods=(POST,)),
            Rule('/api/configure/import', endpoint=self.check(configure.read_import, empty=False), methods=(GET,)),
            Rule('/api/configure/import', endpoint=self.check(configure.write_import, empty=False, busy=True), methods=(POST,)),
            Rule('/api/configure/constants', endpoint=self.check(configure.read_constants, empty=False), methods=(GET,)),
            Rule('/api/configure/constant', endpoint=self.check(configure.write_constant, empty=False), methods=(PUT,)),
            Rule('/api/configure/delete-constant', endpoint=self.check(configure.delete_constant, empty=False), methods=(PUT,)),

            Rule('/api/diary/neighbour-activities/<date>', endpoint=diary.read_neighbour_activities, methods=(GET,)),
            Rule('/api/diary/active-days/<month>', endpoint=diary.read_active_days, methods=(GET,)),
            Rule('/api/diary/active-months/<year>', endpoint=diary.read_active_months, methods=(GET,)),
            Rule('/api/diary/statistics', endpoint=self.check(diary.write_statistics), methods=(PUT,)),
            Rule('/api/diary/latest', endpoint=diary.read_latest, methods=(GET,)),
            Rule('/api/diary/<date>', endpoint=self.check(diary.read_diary), methods=(GET,)),

            Rule('/api/search/activity/<query>', endpoint=search.query_activity, methods=(GET,)),
            Rule('/api/search/activity-terms', endpoint=search.read_activity_terms, methods=(GET,)),

            Rule('/api/jupyter/parameters', endpoint=self.check(jupyter.read_parameters), methods=(GET,)),
            Rule('/api/jupyter/<template>', endpoint=jupyter, methods=(GET,)),

            Rule('/api/kit/edit', endpoint=self.check(kit.read_edit, empty=False), methods=(GET,)),
            Rule('/api/kit/retire-item', endpoint=self.check(kit.write_retire_item, empty=False), methods=(PUT,)),
            Rule('/api/kit/replace-model', endpoint=self.check(kit.write_replace_model, empty=False), methods=(PUT,)),
            Rule('/api/kit/add-component', endpoint=self.check(kit.write_add_component, empty=False), methods=(PUT,)),
            Rule('/api/kit/add-group', endpoint=self.check(kit.write_add_group, empty=False), methods=(PUT,)),
            Rule('/api/kit/items', endpoint=self.check(kit.read_items, empty=False), methods=(GET,)),
            Rule('/api/kit/statistics', endpoint=self.check(kit.read_statistics, empty=False), methods=(GET,)),
            Rule('/api/kit/<date>', endpoint=self.check(kit.read_snapshot, empty=False), methods=(GET,)),

            Rule('/api/thumbnail/<int:activity>', endpoint=thumbnail, methods=(GET,)),
            Rule('/api/thumbnail/<int:activity>/<int:sector>', endpoint=thumbnail, methods=(GET,)),

            Rule('/api/route/latlon/activity/<int:activity>', endpoint=self.check(route.read_activity_latlon, empty=False), methods=(GET,)),
            Rule('/api/route/latlon/sector/<int:sector>', endpoint=self.check(route.read_sector_latlon, empty=False), methods=(GET,)),

            Rule('/api/sector', endpoint=self.check(sector.create_sector, empty=False), methods=(POST,)),
            Rule('/api/sector/<int:sector>', endpoint=self.check(sector.read_sector_journals, empty=False), methods=(GET,)),

            Rule('/api/sparkline/<int:statistic>', endpoint=sparkline, methods=(GET,)),
            Rule('/api/sparkline/<int:statistic>/<int:sector>', endpoint=sparkline, methods=(GET,)),
            Rule('/api/sparkline/<int:statistic>/<int:sector>/<int:activity>', endpoint=sparkline, methods=(GET,)),
            Rule('/api/isparkline/<int:statistic>', endpoint=sparkline, methods=(GET,), defaults={'invert': True}),
            Rule('/api/isparkline/<int:statistic>/<int:sector>', endpoint=sparkline, methods=(GET,), defaults={'invert': True}),
            Rule('/api/isparkline/<int:statistic>/<int:sector>/<int:activity>', endpoint=sparkline, methods=(GET,), defaults={'invert': True}),

            Rule('/api/statistics/plottable', endpoint=self.check(statistics.read_all_plottable), methods=(GET,)),
            Rule('/api/statistics/by-date/<name>', endpoint=self.check(statistics.read_values), methods=(GET,)),

            Rule('/api/static/<path:path>', endpoint=static, methods=(GET,)),

            Rule('/api/upload', endpoint=self.check(upload, empty=False), methods=(PUT,)),

            Rule('/api/warnings', endpoint=self.read_warnings, methods=(GET,)),
            Rule('/api/busy', endpoint=self.read_busy, methods=(GET,)),
            Rule('/api/<path:_>', endpoint=error(BadRequest)),

            # ignore path and serve index.html
            Rule('/<path:_>', defaults={'path': 'index.html'}, endpoint=static, methods=(GET,)),
            Rule('/', defaults={'path': 'index.html'}, endpoint=static, methods=(GET,))
        ])

        self.__configure = configure

    def dispatch_request(self, request):
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            values.pop('_', None)
            if self.__config.db:
                with self.__config.db.session_context() as s:
                    return endpoint(request, s, **values)
            else:
                return endpoint(request, None, **values)
        except HTTPException as e:
            return e

    def wsgi_app(self, environ, start_response):
        request = JSONRequest(environ)
        response = self.dispatch_request(request)
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def read_warnings(self, request, s):
        warnings = []
        if self.__warn_data:
            warnings.append({'title': 'Your Data Are Not Stored Permanently',
                             'text': 'The default Docker configuration stores your data within the same '
                                     'container that runs the code.  If you update / delete / prune the '
                                     'container you will lose your data.'})
        if self.__warn_secure:
            warnings.append({'title': 'This System Is Not Secured For External Use',
                             'text': 'Choochoo should not be deployed on a public server.  '
                                     'It is intended only for local, personal use.'})
        return JsonResponse({DATA: warnings})

    def read_busy(self, request, s):
        return JsonResponse({DATA: self.__configure.is_busy()})

    def check(self, handler, config=True, empty=True, busy=False):

        def wrapper(request, s, *args, **kargs):
            if config:
                if not self.__configure.is_configured():
                    log.debug(f'Redirect (not configured)')
                    return JsonResponse({REDIRECT: '/configure/initial'})
                # if we don't care about config we certainly don't care about data
                if s and empty:
                    if self.__configure.is_empty(s):
                        log.debug(f'Redirect (no data)')
                        return JsonResponse({REDIRECT: '/upload'})
                # todo - should we test for if s here?
                if s and busy and self.__config.exists_any_process():
                    return JsonResponse({REDIRECT: '.'})  # todo - does this work?
            data = handler(request, s, *args, **kargs)
            msg = f'Returning data: {data}'
            if len(msg) > MAX_MSG:
                msg = msg[:MAX_MSG-20] + ' ... ' + msg[-10:]
            log.debug(msg)
            return JsonResponse({DATA: data})

        return wrapper




from ..model import DB
from ...sql import StatisticJournal, Source


def str_or_int_or_none(x):
    return x is None or isinstance(x, str) or isinstance(x, int)


def rewrite_db(model):
    if isinstance(model, list):
        return [rewrite_db(m) for m in model]
    else:
        if DB in model:
            db = model[DB]
            if hasattr(db, 'id'):
                model[DB] = db.id
            elif isinstance(db, str):  # images have urls (no they don't - is this still used?)
                model[DB] = db
            elif isinstance(db, int):  # direct id
                model[DB] = db
            else:  # links have tuples
                model[DB] = [x if str_or_int_or_none(x) else x.id for x in db]
        return model



from logging import getLogger

from ch2.names import SPACE, simple_name
from tests import LogTestCase

log = getLogger(__name__)


class TestName(LogTestCase):

    def test_tokenzie(self):
        self.assertEqual(simple_name('ABC 123 *^%'), 'abc-123-%')  # support for like
        self.assertEqual(simple_name('****'), SPACE)
        self.assertEqual(simple_name('123'), '-123')
        self.assertEqual(simple_name('Fitness 7d'), 'fitness-7d')



from sqlalchemy_utils import database_exists

from .log import log_current_exception


def database_really_exists(uri):
    try:
        return database_exists(uri)
    except Exception:
        log_current_exception(traceback=False)
        return False



from logging import getLogger

import numpy as np
from scipy.interpolate import UnivariateSpline

from .frame import present
from ..names import N

log = getLogger(__name__)


def smooth_elevation(df, smooth=4):
    if not present(df, N.ELEVATION):
        log.debug(f'Smoothing {N.SRTM1_ELEVATION} to get {N.ELEVATION}')
        unique = df.loc[~df[N.DISTANCE].isna() & ~df[N.SRTM1_ELEVATION].isna(),
                        [N.DISTANCE, N.SRTM1_ELEVATION]].drop_duplicates(N.DISTANCE)
        # the smoothing factor is from eyeballing results only.  maybe it should be a parameter.
        # it seems better to smooth along the route rather that smooth the terrain model since
        # 1 - we expect the route to be smoother than the terrain in general (roads / tracks)
        # 2 - smoothing the 2d terrain is difficult to control and can give spikes
        # 3 - we better handle errors from mismatches between terrain model and position
        #     (think hairpin bends going up a mountainside)
        # the main drawbacks are
        # 1 - speed on loading
        # 2 - no guarantee of consistency between routes (or even on the same routine retracing a path)
        spline = UnivariateSpline(unique[N.DISTANCE], unique[N.SRTM1_ELEVATION], s=len(unique) * smooth)
        df[N.ELEVATION] = spline(df[N.DISTANCE])
        df[N.GRADE] = (spline.derivative()(df[N.DISTANCE]) / 10)  # distance in km, but percentage
        df[N.GRADE] = df[N.GRADE].rolling(5, center=True).median().ffill().bfill()
        # avoid extrapolation / interpolation
        df.loc[df[N.SRTM1_ELEVATION].isna(), [N.ELEVATION]] = None
    return df


def add_gradient(df):
    # not used above, but used for barometer based data
    df[N.GRADE] = df[N.ELEVATION].rolling(3, center=True).mean().diff() / (10 * df[N.DISTANCE].diff())
    df[N.GRADE].replace([np.inf, -np.inf], np.nan, inplace=True)



from contextlib import contextmanager
from logging import getLogger, INFO
from urllib.parse import urlsplit

from sqlalchemy import text
from uritools import uriunsplit, urisplit

from .log import log_current_exception
from .names import ADMIN_USER, ADMIN_PASSWD, URI, valid_name, assert_name, USER, PREVIOUS, PASSWD

log = getLogger(__name__)


@contextmanager
def with_log(msg):
    log.warning(msg)
    try:
        yield
        log.info(msg.replace('ing ', 'ed '))
    except:
        log_current_exception(exception_level=INFO, first=True)
        msg = 'Error ' + msg[0].lower() + msg[1:]
        raise Exception(msg)


def get_cnxn(config, **kargs):
    return config.get_database(user=config.args[ADMIN_USER], passwd=config.args[ADMIN_PASSWD], **kargs) \
        .engine.connect()


def get_postgres_cnxn(config):
    return get_cnxn(config, uri=uriunsplit(urisplit(config.args[URI])._replace(path='/postgres')))


def print_query(cnxn, query, extended=False):
    log.debug(query)
    for row in cnxn.execute(text(query)).fetchall():
        name = row[0]
        if valid_name(name, extended=extended):
            print(name)


def list_users(config):
    print_query(get_postgres_cnxn(config), 'select rolname from pg_roles')


def list_schemas(config):
    print_query(get_cnxn(config), 'select nspname from pg_namespace', extended=True)


def list_databases(config):
    print_query(get_postgres_cnxn(config), 'select datname from pg_database')


def quote(cnxn, name):
    # not really needed as we have already validated the name, but let's be as secure as possible
    return cnxn.engine.dialect.identifier_preparer.quote_identifier(name)


def execute(cnxn, stmt, **kargs):
    log.debug(f'Executing {stmt} with {kargs}')
    return cnxn.execute(text(stmt), **kargs)


def remove(cnxn, part, name, extra='', extended=False):
    name = assert_name(name, extended=extended)
    with with_log(f'Removing {part} {name}'):
        stmt = f'drop {part} {quote(cnxn, name)}' + extra
        execute(cnxn, stmt)


def remove_user(config):
    remove(get_postgres_cnxn(config), 'role', config.args[USER])


def remove_database(config):
    remove(get_postgres_cnxn(config).execution_options(isolation_level='AUTOCOMMIT'),
           'database',
           urlsplit(config.args._format(URI)).path[1:])


def test_schema(cnxn, schema):
    return bool(execute(cnxn, 'select 1 from pg_namespace where nspname = :schema', schema=schema).first())


def backup_schema(config):
    user = config.args[USER]
    assert_name(user)
    previous = user + ':' + PREVIOUS
    cnxn = get_cnxn(config)
    if test_schema(cnxn, previous):
        remove(cnxn, 'schema', previous, ' cascade', extended=True)
    add_schema(config, schema=previous, extended=True, set_search_path=False, postgis_writer=False)
    q_user = quote(cnxn, user)
    q_previous = quote(cnxn, previous)
    # https://wiki.postgresql.org/wiki/Clone_schema
    with with_log(f'Copying tables to {previous}'):
        for row1 in execute(cnxn, "select table_name from information_schema.tables "
                                  "where table_schema = :schema and table_type = 'BASE TABLE'",
                            schema=user).fetchall():
            table = row1[0]
            src = f'{q_user}.{quote(cnxn, table)}'
            dst = f'{q_previous}.{quote(cnxn, table)}'
            log.info(f'Creating {dst}')
            execute(cnxn, f'create table {dst} ' 
                          f'(like {src} including constraints including indexes including defaults)')
            log.info(f'Copying {src} to {dst}')
            execute(cnxn, f'insert into {dst} (select * from {src})')
        # for row1 in execute(cnxn, 'select table_name from information_schema.tables WHERE table_schema = :schema',
        #                     schema=user).fetchall():
        #     table = row1[0]
        #     src = f'{q_user}.{quote(cnxn, table)}'
        #     dst = f'{q_previous}.{quote(cnxn, table)}'
        #     log.info(f'Setting foreign key constraints on {dst}')
        #     for row2 in execute(cnxn, f'select pg_get_constraintdef(oid), conname from pg_constraint '
        #                               f'where contype=\'f\' and conrelid = \'{src}\'::regclass'):
        #         key, name = row2[0], row2[1]
        #         key = key.replace(user, previous)
        #         stmt = f'alter table {dst} add constraint {name} {key}'
        #         log.debug(key)
        #         execute(cnxn, stmt)


def truncate_tables(config):
    user = config.args[USER]
    cnxn = get_cnxn(config)
    q_user = quote(cnxn, user)
    with with_log(f'Truncating tables in {user}'):
        for row1 in execute(cnxn, "select table_name from information_schema.tables "
                                  "where table_schema = :schema and table_type = 'BASE TABLE'",
                            schema=user).fetchall():
            table = row1[0]
            src = f'{q_user}.{quote(cnxn, table)}'
            log.info(f'Truncating {src}')
            execute(cnxn, f'truncate table {src} cascade')


def remove_schema(config, previous=True):
    if previous and config.args[PREVIOUS]:
        backup_schema(config)
    truncate_tables(config)
    remove(get_cnxn(config), 'schema', config.args[USER], ' cascade')


def add(cnxn, part, name, stmt, extended=False, **kargs):
    assert_name(name, extended=extended)
    with with_log(f'Adding {part} {name}'):
        stmt = stmt.format(name=quote(cnxn, name))
        execute(cnxn, stmt, **kargs)


def add_user(config):
    cnxn = get_postgres_cnxn(config)
    add(cnxn, 'user', config.args[USER],
        # note that :xxx invokes sqlalchemy's substitution of parameters in text()
        'create role {name} with login password :passwd',
        passwd=config.args[PASSWD])


def _create_role(cnxn, name):
    # https://stackoverflow.com/a/49858797
    execute(cnxn, f'''
do $$
begin
  create role {name} inherit;
  exception when duplicate_object then
    raise notice 'not creating {name} - already exists';
end
$$;
''')


def add_database(config):
    cnxn = get_postgres_cnxn(config).execution_options(isolation_level='AUTOCOMMIT')
    add(cnxn, 'database', urlsplit(config.args._format(URI)).path[1:],
        f'create database {{name}} with owner {quote(cnxn, config.args[USER])}')
    cnxn = get_cnxn(config)
    # https://dba.stackexchange.com/a/37373
    execute(cnxn, 'create extension if not exists btree_gist')
    execute(cnxn, 'create extension if not exists postgis')
    _create_role(cnxn, 'postgis_reader')
    execute(cnxn, 'grant select on geometry_columns to postgis_reader')
    execute(cnxn, 'grant select on geography_columns to postgis_reader')
    execute(cnxn, 'grant select on spatial_ref_sys to postgis_reader')
    _create_role(cnxn, 'postgis_writer')
    execute(cnxn, 'grant postgis_reader to postgis_writer')
    execute(cnxn, 'grant insert, update, delete on spatial_ref_sys to postgis_writer')


def set(cnxn, part, schema, user, stmt, extended=False):
    assert_name(schema, extended=extended)
    assert_name(user)
    with with_log(f'Setting {part} on {schema} for {user}'):
        stmt = stmt.format(schema=quote(cnxn, schema), user=quote(cnxn, user))
        log.debug(stmt)
        cnxn.execute(text(stmt))


def add_schema(config, schema=None, extended=False, set_search_path=True, postgis_writer=True):
    user = config.args[USER]
    assert_name(user)
    schema = schema or user
    assert_name(schema, extended=extended)
    cnxn = get_cnxn(config)
    add(cnxn, 'schema', schema, f'create schema {{name}} authorization {quote(cnxn, user)}',
        extended=extended)
    set(cnxn, 'usage', schema, user, 'grant usage on schema {schema} to {user}',
        extended=extended)
    set(cnxn, 'permissions', schema, user,
        'grant insert, select, update, delete on all tables in schema {schema} to {user}',
        extended=extended)
    set(cnxn, 'permissions', schema, user,
        'alter default privileges in schema {schema} grant insert, select, update, delete on tables to {user}',
        extended=extended)
    if set_search_path:
        set(cnxn, 'search_path', schema, user, 'alter role {user} set search_path to {schema}, public',
            extended=extended)
    if postgis_writer:
        add(cnxn, 'postgis_writer', user, 'grant postgis_writer to {name}')



from logging import getLogger

from werkzeug import Request
from werkzeug.exceptions import BadRequest

log = getLogger(__name__)


class CsrfCheck:

    # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        request = Request(environ)
        if request.method == 'GET':
            log.debug('No CSRF check for GET')
        else:
            if 'CsrfCheck' in request.headers:
                log.debug('Found CsrfCheck')
            else:
                log.error(f'No CsrfCheck header for {request.method}')
                raise BadRequest()
        return self.app(environ, start_response)



from logging import getLogger

from .args import DATE
from ..sql import ActivityJournal, FileScan

log = getLogger(__name__)


def delete(config):
    '''
## delete

    > ch2 delete 2020-01-01

Delete an activity from the database.
    '''
    with config.db.session_context() as s:
        activity = ActivityJournal.at(s, config.args[DATE])
        log.warning('Deleting file scan')
        s.query(FileScan).filter(FileScan.file_hash_id == activity.file_hash_id).delete()
        log.warning('Deleting activity')
        s.delete(activity)
        log.info('Done')



from re import sub, ASCII

from .date import to_time

ADMIN_USER = 'admin-user'
ADMIN_PASSWD = 'admin-passwd'
BASE = 'base'
BIND = 'bind'
COLOR = 'color'
COLOUR = 'colour'
COMMAND = 'command'
DARK = 'dark'
DB = 'db'
L = 'l'
LIGHT = 'light'
LIST = 'list'
OFF = 'off'
P = 'p'
PASSWD = 'passwd'
PORT = 'port'
POSTGRESQL = 'postgresql'
PREVIOUS = 'previous'
SQLITE = 'sqlite'
U = 'u'
URI = 'uri'
USER = 'user'
V = 'v'
VERBOSITY = 'verbosity'
VERSION = 'version'
WEB = 'web'

UNDEF = object()
TIME_ZERO = to_time(0.0)


def assert_name(name, extended=False):
    if not valid_name(name, extended=extended):
        raise Exception(f'Bad name: {name}')
    return name


def valid_name(name, extended=False):
    # this relies on '_' not being a valid identifier to exclude various system tables
    if extended:
        clean = sub(r'[^-A-Za-z0-9:]', '', name, flags=ASCII)
    else:
        clean = sub(r'[^-A-Za-z0-9]', '', name, flags=ASCII)
    if clean != name: return False
    if name.startswith('template'): return False
    if name in ('postgres', 'admin', 'public'): return False
    return True




import datetime as dt
from collections import namedtuple
from logging import getLogger

from sqlalchemy import desc

from .elevation import ElevationCalculator
from .impulse import ImpulseCalculator
from .power import PowerCalculator
from .response import ResponseCalculator
from .utils import ActivityJournalProcessCalculator, DataFrameCalculatorMixin
from ..pipeline import OwnerInMixin, LoaderMixin
from ..read.activity import ActivityReader
from ...data import Statistics
from ...data.activity import active_stats, times_for_distance, hrz_stats, max_med_stats, max_mean_stats, \
    direction_stats, copy_times, round_km, MAX_MINUTES
from ...data.response import response_stats, DIGITS
from ...lib.data import safe_dict
from ...names import N, T, S, U, SPACE
from ...sql import ActivityJournal, StatisticJournal, StatisticJournalType, StatisticName, Sector
from ...sql.tables.sector import SectorJournal, SectorType

log = getLogger(__name__)


class ActivityCalculator(LoaderMixin, OwnerInMixin, DataFrameCalculatorMixin,
                         ActivityJournalProcessCalculator):

    def __init__(self, *args, response_prefix=None, **kargs):
        self.response_prefix = response_prefix
        super().__init__(*args, add_serial=False, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self._provides(s, T.START, StatisticJournalType.TIMESTAMP, None, None,
                       'The start time for the activity.')
        self._provides(s, T.FINISH, StatisticJournalType.TIMESTAMP, None, None,
                       'The finish time for the activity.')
        self._provides(s, T.TIME, StatisticJournalType.FLOAT, U.S, S.join(S.MAX, S.SUM, S.MSR),
                       'The total duration of the activity.')
        self._provides(s, T.ACTIVE_DISTANCE, StatisticJournalType.FLOAT, U.KM, S.join(S.MAX, S.CNT, S.SUM, S.MSR),
                       'The total distance travelled while active (ie not paused).')
        self._provides(s, T.ACTIVE_TIME, StatisticJournalType.FLOAT, U.S, S.join(S.MAX, S.SUM, S.MSR),
                       'The total time while active (ie not paused).')
        self._provides(s, T.ACTIVE_SPEED, StatisticJournalType.FLOAT, U.KMH, S.join(S.MAX, S.AVG, S.MSR),
                       'The average speed while active (ie not paused).')
        self._provides(s, T.MEAN_POWER_ESTIMATE, StatisticJournalType.FLOAT, U.W, S.join(S.MAX, S.AVG, S.MSR),
                       'The average estimated power.')
        self._provides(s, T.DIRECTION, StatisticJournalType.FLOAT, U.DEG, None,
                       'The angular direction (clockwise from North) of the mid-point of the activity relative to the start.')
        self._provides(s, T.ASPECT_RATIO, StatisticJournalType.FLOAT, None, None,
                       'The relative extent of the activity along and across the {T.DIRECTION}.')
        self._provides(s, T.MIN_KM_TIME, StatisticJournalType.FLOAT, U.S, S.join(S.MIN, S.MSR),
                       'The shortest time required to cover the given distance.', values=round_km())
        self._provides(s, T.MED_KM_TIME, StatisticJournalType.FLOAT, U.S, S.join(S.MIN, S.MSR),
                       'The median (typical) time required to cover the given distance.', values=round_km())
        self._provides(s, T.PERCENT_IN_Z, StatisticJournalType.FLOAT, U.PC, None,
                       'The percentage of time in the given HR zone.', values=range(1, 7))
        self._provides(s, T.TIME_IN_Z, StatisticJournalType.FLOAT, U.S, None,
                       'The total time in the given HR zone.', values=range(1, 7))
        self._provides(s, T.MAX_MED_HR_M, StatisticJournalType.FLOAT, U.BPM, S.join(S.MAX, S.MSR),
                       'The highest median HR in the given interval.', values=MAX_MINUTES)
        self._provides(s, T.MAX_MEAN_PE_M, StatisticJournalType.FLOAT, U.W, S.join(S.MAX, S.MSR),
                       'The highest average power estimate in the given interval.', values=MAX_MINUTES)
        self._provides(s, T.TOTAL_CLIMB, StatisticJournalType.FLOAT, U.M, S.join(S.MAX, S.MSR),
                       'The total height climbed in the detected climbs (only).')
        # these are complicated :( because exact names are calculated elsewhere
        for statistic_name in s.query(StatisticName). \
                filter(StatisticName.name.like('%' + N.FITNESS_ANY),
                       StatisticName.owner == ResponseCalculator).all():
            self._provides(s, T._delta(statistic_name.title), StatisticJournalType.FLOAT, U.FF, S.join(S.MAX, S.MSR),
                           'The change (over the activity) in the SHRIMP Fitness parameter.')
            days = int(DIGITS.search(statistic_name.name).group(1))
            # there may be a bug here with the name prefix not being propagated
            self._provides(s, T.EARNED_D % days, StatisticJournalType.FLOAT, U.S, S.join(S.MAX, S.MSR),
                           'The time before Fitness returns to the value before the activity.')
            self._provides(s, T.PLATEAU_D % days, StatisticJournalType.FLOAT, U.S, S.join(S.MAX, S.MSR),
                           'The maximum Fitness achieved if this activity was repeated (with the same time gap to the previous).')
        for statistic_name in s.query(StatisticName). \
                filter(StatisticName.name.like('%' + N.FATIGUE_ANY),
                       StatisticName.owner == ResponseCalculator).all():
            self._provides(s, T._delta(statistic_name.title), StatisticJournalType.FLOAT, U.FF, S.join(S.MAX, S.MSR),
                           'The change (over the activity) in the SHRIMP Fatigue parameter.')
            days = int(DIGITS.search(statistic_name.name).group(1))
            # there may be a bug here with the name prefix not being propagated
            self._provides(s, T.RECOVERY_D % days, StatisticJournalType.FLOAT, U.S, S.join(S.MAX, S.MSR),
                           'The time before Fatigue returns to the value before the activity.')

    def _read_dataframe(self, s, ajournal):
        try:
            adf = Statistics(s, activity_journal=ajournal, with_timespan=True). \
                by_name(ActivityReader, N.DISTANCE, N.HEART_RATE, N.SPHERICAL_MERCATOR_X, N.SPHERICAL_MERCATOR_Y). \
                by_name(ElevationCalculator, N.ELEVATION). \
                by_name(ImpulseCalculator, N.HR_ZONE). \
                by_name(PowerCalculator, N.POWER_ESTIMATE).df
            if self.response_prefix:
                start, finish = ajournal.start - dt.timedelta(hours=1), ajournal.finish + dt.timedelta(hours=1)
                fitness = self.response_prefix + SPACE + N.FITNESS_ANY
                fatigue = self.response_prefix + SPACE + N.FATIGUE_ANY
                sdf = Statistics(s, start=start, finish=finish). \
                    by_name(self.owner_in, fitness, fatigue, like=True).with_. \
                    drop_prefix(self.response_prefix).df
            else:
                sdf = None
            climbs = s.query(SectorJournal). \
                join(Sector). \
                filter(SectorJournal.activity_journal == ajournal,
                       Sector.type == SectorType.CLIMB). \
                all()
            prev = s.query(ActivityJournal). \
                filter(ActivityJournal.start < ajournal.start). \
                order_by(desc(ActivityJournal.start)). \
                first()
            delta = (ajournal.start - prev.start).total_seconds() if prev else None
            return adf, sdf, climbs, delta
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise

    def _calculate_stats(self, s, ajournal, data):
        adf, sdf, climbs, delta = data
        stats = {}
        stats.update(copy_times(ajournal))
        stats.update(active_stats(adf))
        stats.update(self.__average_power(s, ajournal, stats[N.ACTIVE_TIME]))
        stats.update(times_for_distance(adf))
        stats.update(hrz_stats(adf))
        stats.update(max_med_stats(adf))
        stats.update(max_mean_stats(adf))
        stats.update(direction_stats(adf))
        if sdf is not None:
            stats.update(response_stats(sdf, delta))
        if climbs:
            stats.update({N.TOTAL_CLIMB: self.__total_climbed(climbs)})
        return data, stats

    def __total_climbed(self, climbs):
        DistanceElevation = namedtuple('DistanceElevation', 'distance,elevation')
        merged = [(DistanceElevation(climb.start_distance, climb.start_elevation),
                   DistanceElevation(climb.finish_distance, climb.finish_elevation))
                  for climb in climbs]
        START, FINISH = 0, 1
        merged = sorted(merged, key=lambda pair: pair[START].distance)
        log.debug(f'Merging climbs: {merged}')
        i = 0
        while i < len(merged) - 1:
            left, right = merged[i], merged[i+1]
            if left[FINISH].distance >= right[FINISH].distance:
                del merged[i+1]
            elif left[FINISH].distance >= right[START].distance:
                merged[i] = (left[START], right[FINISH])
                del merged[i+1]
            else:
                i += 1
        log.debug(f'Merged climbs: {merged}')
        total = sum(finish.elevation - start.elevation for start, finish in merged)
        log.debug(f'Total {total} from merged climbs: {merged}')
        return total

    @safe_dict
    def __average_power(self, s, ajournal, active_time):
        # this doesn't fit nicely anywhere...
        energy = StatisticJournal.at(s, ajournal.start, N.ENERGY_ESTIMATE, PowerCalculator,
                                     ajournal.activity_group)
        if energy and active_time:
            return {N.MEAN_POWER_ESTIMATE: 1000 * energy.value / active_time}
        else:
            return {N.MEAN_POWER_ESTIMATE: 0}

    def _copy_results(self, s, ajournal, loader, data):
        df, stats = data
        for name in stats.keys():
            loader.add_data(name, ajournal, stats[name], ajournal.start)



from json import dumps, loads
from logging import getLogger
from pydoc import locate
from re import compile, IGNORECASE

import pytz
from geoalchemy2 import Geography
from geoalchemy2.shape import from_shape
from shapely.geometry import Point
from sqlalchemy import TypeDecorator, Integer, Text, func, DateTime

from .utils import WGS84_SRID
from ..names import simple_name

log = getLogger(__name__)


CLS_CACHE = {}


class Cls(TypeDecorator):

    impl = Text
    cache_ok = True

    def process_literal_param(self, cls, dialect):
        return long_cls(cls)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return lookup_cls(value)


def long_cls(cls):
    if cls is None:
        return cls
    if not isinstance(cls, str) and not isinstance(cls, type):
        cls = type(cls)
    if isinstance(cls, type):
        cls = cls.__module__ + '.' + cls.__name__
    return cls


def lookup_cls(value):
    # https://stackoverflow.com/a/24815361
    if not value:
        return None
    if value not in CLS_CACHE:
        CLS_CACHE[value] = locate(value)
    if not CLS_CACHE[value]:
        raise Exception('Cannot find %s' % value)
    return CLS_CACHE[value]


class ShortCls(TypeDecorator):

    impl = Text
    cache_ok = True

    def process_literal_param(self, cls, dialect):
        return short_cls(cls)

    process_bind_param = process_literal_param


def short_cls(cls):
    if cls is None:
        return cls
    if not isinstance(cls, str) and not isinstance(cls, type):
        cls = type(cls)
    if isinstance(cls, type):
        cls = cls.__name__
    return cls


class NullText(TypeDecorator):
    '''
    None (NULL) values are converted to 'None'.

    Use for constraint, where NULLs are distinct in UNIQUE constraints.
    '''

    impl = Text
    cache_ok = True

    def process_literal_param(self, value, dialect):
        return str(value)

    process_bind_param = process_literal_param

    coerce_to_is_types = tuple()  # don't want "IS NULL" when comparing with None


class Json(TypeDecorator):

    impl = Text
    cache_ok = True

    def process_literal_param(self, value, dialect):
        return dumps(value)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return loads(value)


class Sched(TypeDecorator):

    impl = Text
    cache_ok = True

    def process_literal_param(self, sched, dialect):
        from ..lib.schedule import Schedule
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        return str(sched)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        from ..lib.schedule import Schedule
        if value is None:
            return None
        return Schedule(value)


class OpenSched(Sched):

    cache_ok = True

    def process_literal_param(self, sched, dialect):
        from ..lib.schedule import Schedule
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        sched.start = None
        sched.finish = None
        return str(sched)

    process_bind_param = process_literal_param


class Sort(TypeDecorator):

    impl = Integer
    cache_ok = True

    def process_literal_param(self, value, dialect):
        if callable(value):
            value = value()
        return value

    process_bind_param = process_literal_param


class Name(TypeDecorator):

    impl = Text
    cache_ok = True

    def process_literal_param(self, value, dialect):
        return simple_name(value)

    process_bind_param = process_literal_param


class QualifiedName(TypeDecorator):

    impl = Text
    cache_ok = True

    def process_literal_param(self, value, dialect):
        if value and ':' in value:
            left, right = value.rsplit(':', 1)
            return simple_name(left) + ':' + simple_name(right)
        else:
            return simple_name(value)

    process_bind_param = process_literal_param


def point(x, y, srid=-1):
    # return from_shape(Point(x, y), srid=srid)
    return f'ST_MakePoint({x}, {y})'


def linestringxyzm(xyzm, type='geography'):
    if xyzm:
        points = [f'ST_MakePoint({x}, {y}, {z}, {m})' for x, y, z, m in xyzm]
        line = f'ST_MakeLine(ARRAY[{", ".join(points)}])'
    else:
        log.warning(f'Empty geo data')
        line = "'LINESTRINGZM EMPTY'::" + type
    return line


def linestringxym(xym, type='geography'):
    if xym:
        points = [f'ST_MakePointM({x}, {y}, {m})' for x, y, m in xym]
        line = f'ST_MakeLine(ARRAY[{", ".join(points)}])'
    else:
        log.warning(f'Empty geo data')
        line = "'LINESTRINGM EMPTY'::" + type
    return line


def linestringxyz(xyz, type='geography'):
    if xyz:
        points = [f'ST_MakePoint({x}, {y}, {z})' for x, y, z in xyz]
        line = f'ST_MakeLine(ARRAY[{", ".join(points)}])'
    else:
        log.warning(f'Empty geo data')
        line = "'LINESTRINGZ EMPTY'::" + type
    return line


def linestringxy(xy, type='geography'):
    if xy:
        points = [f'ST_MakePoint({x}, {y})' for x, y in xy]
        line = f'ST_MakeLine(ARRAY[{", ".join(points)}])'
    else:
        log.warning(f'Empty geo data')
        line = "'LINESTRING EMPTY'::" + type
    return line


NAME = 'name'
TITLE = 'title'
OWNER = 'owner'


def name_and_title(kargs):
    '''
    allow one or the other to be specified, with special treatment for name only to support legacy code.
    '''

    def add_owner(msg):
        if OWNER in kargs: msg += f' (owner {short_cls(kargs[OWNER])})'
        return msg

    if kargs.get(NAME, None):  # not missing and not None
        name = kargs[NAME]
        if not kargs.get(TITLE, None):
            kargs[TITLE] = name
            log.warning(add_owner(f'Setting title from name "{name}"'))
            kargs[NAME] = simple_name(name)
        else:
            if kargs[NAME] != simple_name(name):
                log.warning(add_owner(f'Unusual name ({name}) for title ({kargs[TITLE]})'))
    elif kargs.get(TITLE, None):
        log.debug(f'Setting name from title ({kargs[TITLE]})')
        kargs[NAME] = simple_name(kargs[TITLE])
    else:
        raise Exception(f'Provide {NAME} or {TITLE}')
    return kargs


class UTC(TypeDecorator):

    impl = DateTime(timezone=True)
    cache_ok = True

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return value.replace(tzinfo=pytz.UTC)




from abc import abstractmethod
from logging import getLogger

from sqlalchemy.sql.functions import count

from .loader import Loader
from ..commands.args import LOG, WORKER, DEV, PROCESS, CPROFILE
from ..common.args import mm
from ..common.global_ import global_dev
from ..common.names import BASE, UNDEF
from ..common.names import VERBOSITY, URI
from ..lib.utils import timing
from ..lib.workers import command_root
from ..sql import Pipeline, Interval, PipelineType, StatisticJournal, StatisticName
from ..sql.types import short_cls

log = getLogger(__name__)

CPU_FRACTION = 0.9
MAX_REPEAT = 3


def count_statistics(s):
    return s.query(count(StatisticJournal.id)).scalar()


def run_pipeline(config, type, like=tuple(), worker=None, **extra_kargs):
    with config.db.session_context() as s:
        if not worker:
            if type == PipelineType.PROCESS:
                Interval.clean(s)
        for pipeline in Pipeline.all(s, type, like=like, id=worker):
            kargs = dict(pipeline.kargs)
            kargs.update(extra_kargs)
            msg = f'Ran {short_cls(pipeline.cls)}'
            if 'activity_group' in kargs: msg += f' ({kargs["activity_group"]})'
            log.debug(f'Running {pipeline}({kargs})')
            with timing(msg):
                before = None if id else count_statistics(s)
                pipeline.cls(config, id=pipeline.id, worker=bool(worker), **kargs).run()
                after = None if id else count_statistics(s)
            if before or after:
                log.info(f'{msg}: statistic count {before} -> {after} (change of {after - before})')


class BasePipeline:

    def __init__(self, *args, **kargs):
        if args or kargs:
            log.warning(f'Unused pipeline argument(s) for {short_cls(self)}: {args} {list(kargs.keys())}')

    def _assert(self, name, value):
        if value is None:
            raise Exception(f'Undefined {name} for {short_cls(self)}')
        else:
            return value

    @abstractmethod
    def run(self):
        raise NotImplementedError()


class ProcessPipeline(BasePipeline):
    '''
    When run via ProcessRunner, an instance is created and either:
    * run() is called, in which case all work (worker or not) is done locally.
      in this case, worker should not call startup() and shutdown(), but a "full run" should.
    * startup(), missing(), command_for_missing() (multiple times, invoking worker threads)
      and shutdown() are called in sequence.  The instance is more like a factory in this case.
    In this way startup and shutdown bracket the entire process and are done just once.
    '''

    def __init__(self, config, *args, owner_out=None, worker=None, id=None, cprofile=None,
                 **kargs):
        self.__args = args
        self._config = config
        self.owner_out = owner_out or self  # the future owner of any calculated statistics
        self.worker = worker
        self.id = id
        self.cprofile = cprofile
        dev = mm(DEV) if global_dev() else ''
        self.__ch2 = f'{command_root()} {mm(BASE)} {config.args[BASE]} {dev} {mm(VERBOSITY)} 0'
        super().__init__(**kargs)

    def startup(self):
        with self._config.db.session_context(expire_on_commit=False) as s:
            log.debug(f'Starting up {self}')
            self._startup(s)

    def _startup(self, s):
        pass

    def shutdown(self):
        with self._config.db.session_context() as s:
            log.debug(f'Shutting down {self}')
            self._shutdown(s)

    def _shutdown(self, s):
        pass

    def missing(self):
        '''
        A missing value identities what is to be processed by a worker.  It is typically a file path or
        the start time (local) of an activity.  It is always a string.  It should be quoted if it contains spaces
        or otherwise needs special handling by the shell.
        '''
        with self._config.db.session_context(expire_on_commit=False) as s:
            missing = self._missing(s) or []  # allow None
        log.debug(f'{len(missing)} missing for {self}')
        return missing

    def _missing(self, s):
        # this should return strings
        raise NotImplementedError('_missing(s)')

    def run(self):
        self.startup()
        if self.worker:
            missing = self.__args
        else:
            missing = [missed.strip('"') for missed in self.missing()]
        for missed in missing:
            self._run_one(missed)
        self.shutdown()

    def _run_one(self, missed):
        # this should accept strings
        raise NotImplementedError('_run_one(missed)')

    def command_for_missing(self, pipeline, missing, log_name):
        from .process import fmt_cmd
        cprofile = ''
        if self.cprofile:
            cprofile = ' ' + mm(CPROFILE)
            if self.cprofile[0]:
                cprofile = ' ' + self.cprofile[0]
        cmd = self.__ch2 + f'{cprofile} {mm(LOG)} {log_name} {mm(URI)} {self._config.args._format(URI)} ' \
                           f'{PROCESS} {mm(WORKER)} {pipeline.id} {" ".join(missing)}'
        log.debug(fmt_cmd(cmd))
        return cmd

    def __str__(self):
        return str(short_cls(self.__class__))

    def _provides(self, s, name, type_, units, summary, description, owner=UNDEF, title=None, values=None):
        if owner is UNDEF:
            owner = self.owner_out
        if values:
            if title: raise Exception('Cannot have title with template name')
            for value in values:
                StatisticName.add_if_missing(s, name % value, type_, units, summary, owner,
                                             description=description)
        else:
            StatisticName.add_if_missing(s, name, type_, units, summary, owner, description=description, title=title)


class LoaderMixin:

    def __init__(self, config, *args, batch=True, **kargs):
        super().__init__(config, *args, **kargs)
        self.__batch = batch

    def _get_loader(self, s, add_serial=None, cls=Loader, **kargs):
        if 'owner' not in kargs:
            kargs['owner'] = self.owner_out
        if add_serial is None:
            raise Exception('Select serial use')
        else:
            kargs['add_serial'] = add_serial
        if 'batch' not in kargs:
            kargs['batch'] = self.__batch
            self.__batch = False  # only set once or we get multiple callbacks
        return cls(s, **kargs)


class OwnerInMixin:

    def __init__(self, *args, owner_in=None, **kargs):
        self.owner_in = self._assert('owner_in', owner_in)
        super().__init__(*args, **kargs)



from contextlib import contextmanager
from logging import getLogger

from .database import SystemConstant, Process, Database
from ..commands.args import DB_VERSION, UNDEF
from ..common.config import BaseConfig
from ..common.log import first_line

log = getLogger(__name__)


class Config(BaseConfig):

    def __init__(self, args):
        super().__init__(args, Database, DB_VERSION)

    def _with(self, **kargs):
        return Config(self.args._with(**kargs))

    def get_constant(self, name, none=False):
        with self.db.session_context() as s:
            value = SystemConstant.from_name(s, name, none=none)
            return value

    def set_constant(self, name, value, force=False):
        log.debug(f'Setting {name}={value}')
        with self.db.session_context() as s:
            return SystemConstant.set(s, name, value, force=force)

    def delete_constant(self, name, default=UNDEF):
        log.debug(f'Deleting {name}')
        with self.default(default):
            with self.db.session_context() as s:
                SystemConstant.delete(s, name)

    def get_process(self, owner, pid):
        with self.db.session_context() as s:
            process = s.query(Process).filter(Process.owner == owner, Process.pid == pid).one_or_none()
            if process is None:
                raise Exception(f'Could not find entry for process {pid} (owner {owner})')
            s.expunge(process)
            return process

    def run_process(self, owner, cmd, log_name, constraint=None):
        with self.db.session_context() as s:
            return Process.run(s, owner, cmd, log_name, constraint=constraint)  # todo change order

    def delete_process(self, owner, pid, delta_seconds=3):
        with self.db.session_context() as s:
            Process.delete(s, owner, pid, delta_seconds=delta_seconds)

    def delete_all_processes(self, owner, delta_seconds=3, default=UNDEF, constraint=None):
        with self.default(default):
            with self.db.session_context() as s:
                Process.delete_all(s, owner, delta_seconds=delta_seconds, constraint=constraint)

    def exists_any_process(self, owner=None, excluding=None):
        with self.db.session_context() as s:
            return Process.exists_any(s, owner=owner, excluding=excluding)

    @contextmanager
    def default(self, value=UNDEF):
        try:
            yield
        except Exception as e:
            if value is UNDEF:
                raise
            else:
                log.warning(f'Error (probably missing database): {first_line(e)}')
                return value



from logging import getLogger

from .args import TOPIC, HELP, PROGNAME
from ..common.args import m
from ..common.md import Markdown

log = getLogger(__name__)


def commands(COMMANDS):
    from .args import H
    return '''
Thank-you for using Choochoo.  Please send feedback to andrew@acooke.org

# Commands

* %s

See also `%s %s` for usage, '%s %s CMD` for guidance on a particular command, 
and `%s %s CMD` for usage of that command.

Docs at http://andrewcooke.github.io/choochoo/index''' % (
        '\n* '.join(COMMANDS.keys()), PROGNAME, m(H), PROGNAME, HELP, PROGNAME, m(H))


def help(config):
    '''
## help

    > ch2 help [topic]

Displays help for a particular topic.

### Examples

    > ch2 help help

Displays this information.

    > ch2 help

Lists available topics.
    '''
    from .. import COMMANDS
    args = config.args
    if args[TOPIC] in COMMANDS:
        Markdown().print(COMMANDS[args[TOPIC]].__doc__)
    else:
        Markdown().print(commands(COMMANDS))





from collections import namedtuple
from functools import cmp_to_key
from itertools import groupby
from logging import getLogger

from sqlalchemy import or_

from .frame import linear_resample, present
from ..common.math import is_nan
from ..lib.data import nearest_index, get_index_loc, safe_yield
from ..names import Names as N
from ..sql import StatisticName, StatisticJournal, Source, ActivityJournal
from ..sql.tables.sector import SectorJournal

log = getLogger(__name__)

SCORE = 'Score'
GRADIENT = 'Gradient'

# a climb of 80m is roughly equivalent to a score of 8000 on strava's weird approach -
# https://support.strava.com/hc/en-us/articles/216917057-How-are-Strava-climbs-categorized-For-Rides-
MIN_CLIMB_ELEVATION = 80
MIN_CLIMB_GRADIENT = 3
MAX_CLIMB_GRADIENT = 40
MAX_CLIMB_REVERSAL = 0.1
CLIMB_CATEGORIES = {MIN_CLIMB_ELEVATION: '4', 160: '3', 320: '2', 640: '1', 800: 'HC'}

# conversion of gradient to percent, but correcting for distance in km
PERCENT = 100 / 1000

# trade-off between pure elevation (0) and pure gradient (1)
CLIMB_PHI = 0.7

Climb = namedtuple('Climb', 'phi, min_elevation, min_gradient, max_gradient, max_reversal',
                   defaults=(CLIMB_PHI, MIN_CLIMB_ELEVATION, MIN_CLIMB_GRADIENT,
                             MAX_CLIMB_GRADIENT, MAX_CLIMB_REVERSAL))


@safe_yield
def find_climbs(df, params=Climb()):
    df = df.drop_duplicates(subset=[N.DISTANCE])
    by_dist = df.set_index(df[N.DISTANCE])
    by_dist = linear_resample(by_dist, quantise=False)
    for dlo, dhi in find_climb_distances(by_dist, params=params):
        tlo, thi = nearest_index(df, N.DISTANCE, dlo), nearest_index(df, N.DISTANCE, dhi)
        log.debug(f'Found climb from {tlo} - {thi} ({dlo}km - {dhi}km)')
        up = df[N.ELEVATION].loc[thi] - df[N.ELEVATION].loc[tlo]
        along = df[N.DISTANCE].loc[thi] - df[N.DISTANCE].loc[tlo]
        climb = {N.TIME: thi,
                 N.CLIMB_ELEVATION: up,
                 N.CLIMB_DISTANCE: along,
                 N.CLIMB_TIME: (thi - tlo).total_seconds(),
                 N.CLIMB_GRADIENT: PERCENT * up / along}
        for height in sorted(CLIMB_CATEGORIES.keys()):
            if up >= height:
                climb[N.CLIMB_CATEGORY] = CLIMB_CATEGORIES[height]
            else:
                break
        log.debug(climb)
        yield climb


def find_climb_distances(df, params=Climb()):
    mn, mx = df[N.ELEVATION].min(), df[N.ELEVATION].max()
    if mx - mn > params.min_elevation:
        score, lo, hi = biggest_climb(df, params=params)
        if score:
            a, b, c = split(df, lo, hi)
            yield from find_climb_distances(a, params=params)
            yield from contiguous(b, params=params)
            yield from find_climb_distances(c, params=params)


def split(df, lo, hi, inside=True):
    ilo, ihi = get_index_loc(df, lo), get_index_loc(df, hi)
    if ilo > ihi:
        ilo, ihi = ihi, ilo
    if inside:
        ihi += 1
    else:
        ilo += 1
    return df.iloc[:ilo].copy(), df.iloc[ilo:ihi].copy(), df.iloc[ihi:].copy()


def contiguous(df, params=Climb()):
    up = df[N.ELEVATION].iloc[-1] - df[N.ELEVATION].iloc[0]
    if up >= params.min_elevation:
        down, lo, hi = biggest_reversal(df)
        if down and down > params.max_reversal * up:
            a, b, c = split(df, lo, hi, inside=False)
            yield from contiguous(a, params=params)
            yield from find_climb_distances(b, params=params)
            yield from contiguous(c, params=params)
        else:
            along = df.index[-1] - df.index[0]
            if along and PERCENT * up / along < params.max_gradient:
                yield df.index[0], df.index[-1]


def first_or_none(generator):
    try:
        return next(generator)
    except StopIteration:
        return None


def biggest_reversal(df):
    max_elevation, max_indices, d = 0, (None, None), df.index[1] - df.index[0]
    for offset in range(1, len(df)-1):
        df[N._delta(N.ELEVATION)] = df[N.ELEVATION].diff(-offset)
        if present(df, N._delta(N.ELEVATION)):
            d_elevation = df[N._delta(N.ELEVATION)].dropna().max()
            if d_elevation > max_elevation:
                max_elevation = d_elevation
                hi = df.loc[df[N._delta(N.ELEVATION)] == max_elevation].index[0]  # break ties
                lo = df.index[get_index_loc(df, hi) + offset]
                max_indices = (lo, hi)
    return max_elevation, max_indices[0], max_indices[1]


def biggest_climb(df, params=Climb(), grid=10):
    # returns (score, dlo, dhi))
    # use distances (indices) rather than ilocs because we're subdividing the data
    if len(df) > 100 * grid:
        score, lo, hi = search(df.iloc[::grid].copy(), grid=True)
        if score:
            # need to pass through iloc to extend range
            ilo, ihi = get_index_loc(df, lo), get_index_loc(df, hi)
            return search(df.iloc[max(0, ilo-grid):min(ihi+grid, len(df)-1)].copy(), params=params)
        else:
            return 0, None, None
    else:
        return search(df, params=params)


def search(df, params=Climb(), grid=False):
    # returns (score, dlo, dhi)
    # use distance (indices) rather than ilocs because we're subdividing the data
    max_score, max_indices, d = 0, (None, None), df.index[1] - df.index[0]
    for offset in range(len(df)-1, 0, -1):
        df[N._delta(N.ELEVATION)] = df[N.ELEVATION].diff(offset)
        d_distance = d * offset
        min_elevation = max(params.min_elevation, params.min_gradient * d_distance / PERCENT)
        if df[N._delta(N.ELEVATION)].max() > min_elevation:  # avoid some work
            # factor of 1000 below to convert km to m
            df[SCORE] = df[N._delta(N.ELEVATION)] / ((1000 * d_distance) ** params.phi)
            score = df.loc[df[N._delta(N.ELEVATION)] > min_elevation, SCORE].max()
            if not is_nan(score) and score > max_score:
                max_score = score
                hi = df.loc[df[SCORE] == max_score].index[0]  # arbitrarily pick one if tied (error here w item())
                lo = df.index[get_index_loc(df, hi) - offset]
                if not grid:
                    # step inwards one location from each end
                    # (so that we have some 'extra' to aid with intersections)
                    lo = df.index[get_index_loc(df, hi) - (offset-1)]
                    hi = df.index[get_index_loc(df, hi) - 1]
                max_indices = (lo, hi)
    return max_score, max_indices[0], max_indices[1]


def climbs_for_activity(s, ajournal):

    from ..pipeline.calculate.sector import SectorCalculator
    from ..pipeline.calculate import ActivityCalculator

    total = s.query(StatisticJournal). \
        join(StatisticName).join(Source). \
        filter(StatisticName.name == N.TOTAL_CLIMB,
               StatisticJournal.time == ajournal.start,
               StatisticName.owner == ActivityCalculator,
               Source.activity_group == ajournal.activity_group).order_by(StatisticJournal.time).one_or_none()
    query = s.query(StatisticJournal). \
        join(StatisticName).join(Source). \
        join(SectorJournal, SectorJournal.id == Source.id). \
        filter(or_(StatisticName.name.like(N.CLIMB_ANY),
                   StatisticName.name == N.VERTICAL_POWER),
               StatisticJournal.time >= ajournal.start,
               StatisticJournal.time <= ajournal.finish,
               StatisticName.owner == SectorCalculator,
               Source.activity_group == ajournal.activity_group).order_by(StatisticJournal.time)
    sjournals = query.all()

    def make_climb(sjournals):
        sjournals = list(sjournals)
        climb = {sjournal.statistic_name.name: sjournal for sjournal in sjournals}
        climb['start-distance'] = sjournals[0].source.start_distance
        climb['sector-journal-id'] = sjournals[0].source.id
        climb['sector-id'] = sjournals[0].source.sector_id
        return climb

    def cmp_climb(a, b):
        # if start distances are similar, order by elevation
        if abs(a['start-distance'] - b['start-distance']) / (a['start-distance'] + b['start-distance']) > 0.03:
            return a['start-distance'] - b['start-distance']
        else:
            return a[N.CLIMB_ELEVATION].value - b[N.CLIMB_ELEVATION].value

    return total, sorted((make_climb(grouped)
                          for _, grouped in groupby(sjournals, key=lambda sjournal: sjournal.time)),
                         key=cmp_to_key(cmp_climb))


def climb_sources(s, activity_journal, activity_group=None):
    if not isinstance(activity_journal, Source):
        activity_journal = ActivityJournal.at(s, activity_journal, activity_group=activity_group)
    return s.query(SectorJournal).filter(SectorJournal.activity_journal_id == activity_journal.id).all()



from logging import getLogger

from sqlalchemy.sql.functions import count

from ch2.commands.args import V, DEV, FORCE, bootstrap_db
from ch2.commands.constants import constants
from ch2.common.args import mm, m
from ch2.config.profiles.default import default
from ch2.sql.tables.constant import Constant
from tests import LogTestCase, random_test_user

log = getLogger(__name__)


class TestConstant(LogTestCase):

    def test_constant(self):
        user = random_test_user()
        config = bootstrap_db(user, m(V), '5')
        bootstrap_db(user, m(V), '5', mm(DEV), configurator=default)
        with config.db.session_context() as s:
            n = s.query(count(Constant.id)).scalar()
            self.assertEqual(n, 14)
        config = bootstrap_db(user, m(V), '5', 'constants', 'set', 'fthr:%', '154', mm(FORCE))
        constants(config)
        with config.db.session_context() as s:
            n = s.query(count(Constant.id)).scalar()
            self.assertEqual(n, 14)
            # todo - maybe test for value?
            # todo - now that this is defined anyway, change the test?



from logging import getLogger, NullHandler
from sys import version_info, exit

from .commands.sparkline import sparkline
from .common.args import NamespaceWithVariables
from .common.global_ import set_global_dev
from .common.md import Markdown
from .common.names import COLOR, BASE, DB

getLogger('bokeh').addHandler(NullHandler())
getLogger('tornado').addHandler(NullHandler())


from .commands.args import COMMAND, make_parser, PROGNAME, HELP, DEV, DIARY, FIT, \
    PACKAGE_FIT_PROFILE, ACTIVITIES, NO_OP, DATABASE, CONSTANTS, SHOW_SCHEDULE, MONITOR, GARMIN, \
    UNLOCK, DUMP, FIX_FIT, CH2_VERSION, JUPYTER, KIT, WEB, IMPORT, THUMBNAIL, CHECK, SEARCH, VALIDATE, \
    DB_VERSION, UPLOAD, PROCESS, DELETE, SPARKLINE
from .commands.process import process
from .commands.upload import upload
from .commands.constants import constants
from .commands.delete import delete
from .commands.validate import validate
from .commands.db import db
from .commands.fit import fit
from .commands.fix_fit import fix_fit
from .commands.help import help
from .commands.import_ import import_
from .commands.kit import kit
from .commands.package_fit_profile import package_fit_profile
from .commands.search import search
from .commands.show_schedule import show_schedule
from .commands.thumbnail import thumbnail
from .commands.web import web
from .lib.log import make_log_from_args
from .sql.database import SystemConstant
from .lib.cprofile import profile
from .sql.config import Config

log = getLogger(__name__)


def no_op(args, system, db):
    '''
## no-op

This is used internally when accessing data in Jupyter or configuring the system
at the command line.
    '''
    pass


COMMANDS = {CONSTANTS: constants,
            DB: db,
            DELETE: delete,
            FIT: fit,
            FIX_FIT: fix_fit,
            HELP: help,
            IMPORT: import_,
            KIT: kit,
            NO_OP: no_op,
            PACKAGE_FIT_PROFILE: package_fit_profile,
            PROCESS: process,
            SEARCH: search,
            SHOW_SCHEDULE: show_schedule,
            SPARKLINE: sparkline,
            THUMBNAIL: thumbnail,
            UPLOAD: upload,
            VALIDATE: validate,
            WEB: web
            }


def args_and_command():
    parser = make_parser()
    ns = parser.parse_args()
    command_name = ns.command if hasattr(ns, COMMAND) else None
    command = COMMANDS[command_name] if command_name in COMMANDS else None
    if command_name == NO_OP: ns.verbose = 0
    args = NamespaceWithVariables._from_ns(ns, PROGNAME, DB_VERSION)
    return args, command, command_name


def versions():
    log.info('Version %s' % CH2_VERSION)
    if version_info < (3, 7):
        raise Exception('Please user Python 3.7 or more recent')


def main():
    versions()
    args, command, command_name = args_and_command()
    set_global_dev(args[DEV])
    make_log_from_args(args)
    config = Config(args)
    with profile(args):
        try:
            if not command:
                log.debug('If you are seeing the "No command given" error during development ' +
                          'you may have forgotten to set the command name via `set_defaults()`.')
                raise Exception('No command given (try `ch2 help`)')
            elif command_name not in (DB, PACKAGE_FIT_PROFILE, HELP, IMPORT, FIT):
                db = None
                try:
                    db = config.db if command_name not in (PACKAGE_FIT_PROFILE, HELP) else None
                except Exception as e:
                    log.warning(e)
                if not db:
                    refuse_until_configured(command_name, False)
                elif db.no_data():
                    refuse_until_configured(command_name, True)
            command(config)
        except KeyboardInterrupt:
            log.critical('User abort')
            exit(1)
        except Exception as e:
            log.critical(e, exc_info=True)
            log.info('See `%s %s` for available commands.' % (PROGNAME, HELP))
            log.info('Docs at http://andrewcooke.github.io/choochoo')
            if not args or args[DEV]:
                raise
            exit(2)


def refuse_until_configured(command_name, uri):
    Markdown().print(f'''
Welcome to Choochoo.

You must configure the database before use (no {"schema" if uri else "uri"}).

Please use the {PROGNAME} {DB} command.
''')
    if command_name != WEB:
        exit(3)



from hashlib import md5
from logging import getLogger
from os import makedirs
from os.path import dirname, normpath, expanduser
from pathlib import Path
from shutil import get_terminal_size

log = getLogger(__name__)


def touch(path, with_dirs=False):
    path = clean_path(path)
    if with_dirs:
        dirs = dirname(path)
        makedirs(dirs, exist_ok=True)
    Path(path).touch()


def clean_path(path):
    # don't use realpath here since it messes up inside docker
    return normpath(expanduser(path))


# https://stackoverflow.com/a/3431838
def file_hash(file_path):
    hash = md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash.update(chunk)
    return hash.hexdigest()


def data_hash(data):
    if isinstance(data, str): data = data.encode('utf-8')
    hash = md5()
    hash.update(data)
    return hash.hexdigest()


def terminal_width(width=None):
    width = get_terminal_size()[0] if width is None else width
    if width is None:
        log.warning('No terminal width available')
    else:
        log.debug(f'Terminal width is {width}')
    return width



from collections import namedtuple
from logging import getLogger

# hide from imports of this package
import numpy as _np
import scipy as _sp
from math import pi
from pandas import to_numeric

import ch2.common.args
from .frame import median_dt
from .lib import fit, inplace_decay
from ..lib.data import tmp_name, safe_first
from ..names import Names as N

log = getLogger(__name__)
RAD_TO_DEG = 180 / pi


def add_differentials(df, max_gap=None):
    return _add_differentials(df, N.SPEED, N.DISTANCE, N.ELEVATION, N.SPEED, N.SPEED_2, max_gap=max_gap)


def _add_differentials(df, speed, *names, max_gap=None):

    # rather than use timespans (old approach) it's more reliable to discard any intervals
    # over a certain time gap.

    speed_2 = N._sqr(speed)
    df[speed_2] = df[speed] ** 2

    tmp = tmp_name()
    df[tmp] = df.index
    df[tmp] = df[tmp].diff().dt.seconds

    for name in names:
        df[N._delta(name)] = df[name].diff()
        df.loc[df[tmp] > max_gap, [N._delta(name)]] = _np.nan

    avg_speed_2 = [(a**2 + a*b + b**2)/3 for a, b in zip(df[speed], df[speed][1:])]
    df[N._avg(speed_2)] = [_np.nan] + avg_speed_2
    if max_gap:
        df.loc[df[tmp] > max_gap, [N._avg(speed_2)]] = _np.nan

    df.drop(columns=[tmp], inplace=True)
    return df


@safe_first
def add_energy_budget(df, m, g=9.8):
    # if DELTA_ELEVATION is +ve we've gone uphill.  so this is the total amount of energy
    # gained in this segment.
    df[N.DELTA_ENERGY] = m * (df[N.DELTA_SPEED_2] / 2 + df[N.DELTA_ELEVATION] * g)
    return df


def add_cda_estimate(df, p=1.225):
    # https://www.cyclingpowerlab.com/CyclingAerodynamics.aspx
    # assume that all energy lost (-ve gain) is due to air resistance.
    df[N.CDA] = -df[N.DELTA_ENERGY] / (p * df[N.AVG_SPEED_2] * df[N.DELTA_DISTANCE] * 0.5)
    return df


def add_crr_estimate(df, m, g=9.8):
    # assume that all energy lost is due to rolling resistance
    df[N.CRR] = -df[N.DELTA_ENERGY] / (df[N.DELTA_DISTANCE] * m * g)
    return df


@safe_first
def add_loss_estimate(df, m, cda=0.45, crr=0, p=1.225, g=9.8):
    # this is the energy spent on air and rolling resistance
    df[N.LOSS] = (cda * p * df[N.AVG_SPEED_2] * 0.5 + crr * m * g) * df[N.DELTA_DISTANCE]
    return df


@safe_first
def add_power_estimate(df):
    # power input must balance the energy budget.
    df[N.VERTICAL_POWER] = df[N.DELTA_ENERGY] / df[N.DELTA_TIME].dt.total_seconds()
    df[N.VERTICAL_POWER].clip(lower=0, inplace=True)
    df[N.POWER_ESTIMATE] = (df[N.DELTA_ENERGY] + df[N.LOSS]) / df[N.DELTA_TIME].dt.total_seconds()
    df[N.POWER_ESTIMATE].clip(lower=0, inplace=True)
    # if N.CADENCE in df.columns:
    #     df.loc[df[N.CADENCE] < 1, [N.POWER_ESTIMATE]] = 0.0
    if any(df[N.POWER_ESTIMATE].isna()):
        df.loc[df[N.POWER_ESTIMATE].isna(), [N.POWER_ESTIMATE]] = 0.0
        # need to coerce because dtype is object
        df[N.POWER_ESTIMATE] = df[N.POWER_ESTIMATE].apply(to_numeric, errors='coerce')
    energy = (df[N.POWER_ESTIMATE].iloc[1:] * df[N.DELTA_TIME].iloc[1:]).cumsum()
    df[N.ENERGY] = 0.0
    df.loc[:, [N.ENERGY]].iloc[1:] = energy
    return df


def evaluate(df, model, quiet=True):
    # used in fitting
    if not quiet: log.debug(f'Evaluating {model}')
    df = add_energy_budget(df, model.bike_weight)
    df = add_loss_estimate(df, model.bike_weight, cda=model.cda, crr=model.crr)
    df = add_power_estimate(df)
    return df



from logging import getLogger

from .. import DEV
from ..commands.args import LOG
from ..common.args import mm
from ..common.global_ import global_dev
from ..common.names import VERBOSITY, BASE, URI
from ..lib.workers import command_root

log = getLogger(__name__)


def run(config, cmd, owner, log_name):
    full_cmd = f'{command_root()} {mm(VERBOSITY)} 0 {mm(BASE)} {config.args[BASE]} ' \
               f'{mm(URI)} {config.args._format(URI)} {mm(LOG)} {log_name}'
    if global_dev(): full_cmd += f' {mm(DEV)}'
    full_cmd += f' {cmd}'
    log.info(f'Starting {full_cmd}')
    return config.run_process(owner, full_cmd, log_name)


def run_and_wait(config, cmd, owner, log_name):
    popen = run(config, cmd, owner, log_name)
    popen.wait()
    config.delete_process(owner, popen.pid)
    return popen.returncode




from sqlalchemy import Column, Text, Integer, ForeignKey, Index, DateTime
from sqlalchemy.orm import relationship, backref

from ..support import Base
from ..types import ShortCls, UTC
from ..utils import add
from ...lib import to_time


class FileHash(Base):
    '''
    This exists so that hash can be used to connect entities (eg activity topics).
    Entries are never deleted from here (but it's created empty on database updates).
    '''

    __tablename__ = 'file_hash'

    id = Column(Integer, primary_key=True)
    hash = Column(Text, nullable=False, index=True, unique=True)

    @classmethod
    def get_or_add(cls, s, hash):
        instance = s.query(FileHash).filter(FileHash.hash == hash).one_or_none()
        if not instance:
            instance = add(s, FileHash(hash=hash))
        return instance


class FileScan(Base):

    __tablename__ = 'file_scan'

    id = Column(Integer, primary_key=True)
    path = Column(Text, nullable=False)
    owner = Column(ShortCls, nullable=False)
    last_scan = Column(UTC, nullable=False)
    file_hash_id = Column(Integer, ForeignKey('file_hash.id'), nullable=False)
    file_hash = relationship('FileHash', backref=backref('file_scan', cascade='all, delete-orphan',
                                                         passive_deletes=True, uselist=False))
    Index('natural_primary_file_scan', path, owner)

    @classmethod
    def add(cls, s, path, owner, hash):
        return add(s, FileScan(path=path, owner=owner, last_scan=to_time(0.0), file_hash=FileHash.get_or_add(s, hash)))

    def __str__(self):
        return self.path



from logging import getLogger

from werkzeug.utils import redirect

from ...commands.args import JUPYTER
from ...jupyter.load import create_notebook
from ...jupyter.utils import get_template
from ...pipeline.display.activity.utils import latest_activity, activity_times_by_group

log = getLogger(__name__)


class Jupyter:

    def __init__(self, config):
        self.__config = config

    def __call__(self, request, s, template):
        log.info(f'Attempting to display template {template}')
        fn, spec = get_template(template)
        args = [request.args[arg] for arg in spec.args]  # order
        log.debug(f'Template args: {args}')
        name = create_notebook(self.__config, fn, args)
        url = f'{self.__config.args[JUPYTER]}/{name}'
        log.debug(f'Redirecting to {url}')
        return redirect(url)

    @staticmethod
    def read_parameters(request, s):
        # odds and sods used to set menus in jupyter URLs
        latest = latest_activity(s)
        times_by_group = activity_times_by_group(s)
        all_activity_times = sorted(sum(times_by_group.values(), []), reverse=True)
        return {'all_activity_times': all_activity_times,
                'activity_times_by_group': times_by_group,
                'latest_activity_group': latest.activity_group.name if latest else None}



from logging import getLogger
from os import getpid
from time import time, sleep

import psutil as ps
from sqlalchemy import Column, Text, Integer, DateTime, distinct

from ..support import Base
from ..types import ShortCls, Name, short_cls, UTC
from ..utils import add
from ...common.date import to_time
from ...lib import now
from ...lib.utils import str_or_none
from ...names import simple_name

log = getLogger(__name__)


class SystemConstant(Base):

    __tablename__ = 'system_constant'

    name = Column(Name, primary_key=True)
    value = Column(Text, nullable=False)

    @classmethod
    def from_name(cls, s, name, none=False):
        instance = s.query(SystemConstant).filter(SystemConstant.name == name).one_or_none()
        if instance is None:
            if not none:
                raise Exception(f'No value for {name}')
            else:
                log.debug(f'Read {name}=None')
                return None
        else:
            log.debug(f'Read {name}={instance.value}')
            return instance.value

    @classmethod
    def set(cls, s, name, value, force=False):
        if force:
            cls.delete(s, name)
        s.add(SystemConstant(name=name, value=value))
        s.commit()
        log.debug(f'Set {name}={value}')
        return value

    @classmethod
    def delete(cls, s, name):
        s.query(SystemConstant).filter(SystemConstant.name == name).delete()
        s.commit()

    TIMEZONE = 'timezone'
    JUPYTER_URL = 'jupyter-url'
    WEB_URL = 'web-url'
    LAST_GARMIN = 'last-garmin'
    DB_VERSION = 'db-version'
    LOG_COLOR = 'log-color'


class Process(Base):

    __tablename__ = 'process'

    id = Column(Integer, primary_key=True)
    owner = Column(ShortCls, nullable=False, index=True)
    constraint = Column(Text)
    pid = Column(Integer, nullable=False, unique=True)
    start = Column(UTC, nullable=False, default=now)
    command = Column(Text, nullable=True)
    log = Column(Text, nullable=True)

    @classmethod
    def run(cls, s, owner, cmd, log_name, constraint=None):
        from ...pipeline.process import fmt_cmd
        popen = ps.Popen(args=cmd, shell=True)
        log.debug(f'Adding command [{fmt_cmd(cmd)}]; pid {popen.pid}')
        s.query(Process).filter(Process.pid == popen.pid).delete(synchronize_session=False)
        s.add(Process(command=cmd, owner=owner, pid=popen.pid, log=log_name,
                      constraint=str_or_none(constraint)))
        s.commit()
        return popen

    @classmethod
    def delete(cls, s, owner, pid, delta_seconds=3):
        # ignore constraint here because we have pid
        process = s.query(Process).filter(Process.owner == owner, Process.pid == pid).one()
        process.__kill(delta_seconds=delta_seconds)
        log.debug(f'Deleting record for process {process.pid}')
        s.delete(process)
        s.commit()

    @classmethod
    def delete_all(cls, s, owner, delta_seconds=3, constraint=None):
        for process in s.query(Process).filter(Process.owner == owner,
                                               Process.constraint == str_or_none(constraint)).all():
            if process.pid == getpid():
                log.debug(f'Not killing self (PID {process.pid})')
            else:
                process.__kill(delta_seconds=delta_seconds)
                log.debug(f'Deleting record for process {process.pid}')
                s.delete(process)
        s.commit()

    @classmethod
    def exists_any(cls, s, owner=None, excluding=None):
        if owner:
            owners = [short_cls(owner)]
        else:
            owners = [row[0] for row in s.query(distinct(Process.owner)).all()]
        owners = set(owners)
        if excluding:
            owners = owners.difference(short_cls(exc) for exc in excluding)
        for owner in owners:
            for process in s.query(Process).filter(Process.owner == owner).all():
                if process.__still_running():
                    return True
            cls.delete_all(s, owner)
        return False

    def __still_running(self, delta_seconds=3):
        return exists(self.pid, self.start, delta_seconds=delta_seconds)

    def __kill(self, delta_seconds=3):
        if self.__still_running(delta_seconds=delta_seconds):
            log.debug(f'Killing process {self.pid}')
            ps.Process(self.pid).kill()


# if we're given a delta time, the start time of the process and database records have to match
# (or it's some other unrelated process because the process counter wrapped round)
# if they don't match we return false because the original process is not running.
def exists(pid, start_time, delta_seconds=3, zombie_seconds=10):
    if not ps.pid_exists(pid):
        return False
    process = ps.Process(pid)
    if process.status() == ps.STATUS_ZOMBIE:
        try:
            log.warning(f'Waiting for zombie {pid}')
            process.wait(zombie_seconds)
        except ps.TimeoutExpired:
            log.debug(f'Timeout waiting for zombie {pid}')
        return False
    elif delta_seconds:
        creation_ok = abs(process.create_time() - start_time.timestamp()) < delta_seconds
        if creation_ok:
            log.debug(f'Process {pid} still exists')
        else:
            log.debug(f'Creation time for process {pid} incorrect ({process.create_time()} / {start_time.timestamp()})')
        return creation_ok
    else:
        log.debug(f'Assuming process {pid} is same as expected')
        return True   # if delta_seconds = 0 don't check, just assume same





from cProfile import Profile
from contextlib import contextmanager
from glob import glob
from logging import getLogger
from os import makedirs, link, getpid
from os.path import exists, isdir, join, abspath

from ..commands.args import START, DURATION, COMMAND, WORKER, CPROFILE
from ..common.date import now, time_to_local_time
from ..common.io import clean_path

log = getLogger(__name__)


@contextmanager
def profile(args):
    use_profile = bool(args[CPROFILE])
    dir, profiler, start = 'profile', None, now()
    if use_profile:
        if args[CPROFILE][0]: dir = args[CPROFILE][0]
        profiler, dir = startup(dir)
    try:
        yield
    finally:
        if use_profile:
            worker = args[WORKER] if WORKER in args else ''
            shutdown(profiler, dir, args[COMMAND], start, worker)


def startup(dir):
    dir = abspath(clean_path(dir))
    log.info(f'Will save profile data to {dir} dir')
    if not exists(dir):
        makedirs(dir, exist_ok=True)
    if not isdir(dir):
        raise Exception(f'{dir} is not a path')
    if glob(f'{dir}/*'):
        log.warning(f'{dir} already contains data')
    makedirs(join(dir, START), exist_ok=True)
    makedirs(join(dir, DURATION), exist_ok=True)
    profiler = Profile()
    profiler.enable()
    return profiler, dir


def shutdown(profiler, dir, command, start, worker):
    profiler.disable()
    duration = int(0.5 + (now() - start).total_seconds())
    duration = f'{duration:06d}'
    start = time_to_local_time(start).replace(" ", "T").replace(':', '-')
    pid = str(getpid())
    path_duration = join(dir, DURATION, f'{duration}.{worker}.{command}.{pid}')
    profiler.dump_stats(path_duration)
    log.warning(f'Saving profile data to {path_duration}')
    path_start = join(dir, START, f'{start}.{worker}.{command}.{pid}')
    link(path_duration, path_start)
    log.warning(f'Saving profile data to {path_start}')




from logging import getLogger

from geoalchemy2 import WKTElement, WKBElement
from geoalchemy2.shape import to_shape
from sqlalchemy import text

from ..utils import ActivityJournalDelegate
from ....diary.model import optional_text, map
from ....sql.utils import SPHM_SRID

log = getLogger(__name__)


class MapDelegate(ActivityJournalDelegate):

    @optional_text('Map')
    def read_journal_date(self, s, ajournal, date):
        q = text(f'''
select st_envelope(st_transform(aj.route_et::geometry, {SPHM_SRID}))
  from activity_journal as aj
 where aj.id = :activity_journal_id
''')
        row = s.connection().execute(q, activity_journal_id=ajournal.id).fetchone()
        envelope = to_shape(WKBElement(row[0]))
        yield map(envelope.bounds, ('activity', str(ajournal.id)))

    def read_interval(self, s, interval):
        # todo?
        return
        yield
